//
//  ***************************************************************************
//
//  bond.cpp
//  (C) 2004 Bosco K. Ho 
//
//  HBondList takes a Protein object, counts the h-bonds, and stores the
//  h-bond information in a list for easy access later on.
//
//  ***************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//


#include "bond.h"


using namespace std;


const double CA_DISTANCE_SQ          =  9.0 * 9.0;
const double Q_SQ                    = -27888.0; // 331.2 * 0.30 * 0.30
const double HBOND_ENERGY_LIMIT      = -9900;    // cal/mol
const double HBOND_ENERGY_HIGH       = -500;     // cal/mol
const double WEAK_HBOND_ENERGY_HIGH  = -1000;     // cal/mol
const double DISULFIDE_DISTANCE_SQ   = 2.5 * 2.5;    //  angstroms



void HBondList::checkBond(Protein* pProtein, int i, int c, int o, int j, int n, int h)
{

  if (pos_dist_squared(pProtein->atom(j, CA).pos, pProtein->atom(i, CA).pos) > CA_DISTANCE_SQ)
    return;

  if (!pProtein->hasAtom(i, c) || !pProtein->hasAtom(i, o)
   || !pProtein->hasAtom(j, n) || !pProtein->hasAtom(j, h))
    return;

  double ho = pos_distance(pProtein->atom(j, h).pos, pProtein->atom(i, o).pos);
  double hc = pos_distance(pProtein->atom(j, h).pos, pProtein->atom(i, c).pos);
  double no = pos_distance(pProtein->atom(j, n).pos, pProtein->atom(i, o).pos);
  double nc = pos_distance(pProtein->atom(j, n).pos, pProtein->atom(i, c).pos);

  // make sure it is a hydrogen bond, i.e. the h is between the donors
  if (ho > nc) return;

  // the kabsch-sanders h-bond function!!!!
  double energy = (long)floor(Q_SQ/ho - Q_SQ/hc + Q_SQ/nc - Q_SQ/no + 0.5);

  if (energy > HBOND_ENERGY_HIGH) return;

  // add this h-bond to the list
  BondStruct bond(i,c,o,j,n,h);
  list.push_back(bond);
}



void HBondList::countDonors(Protein* pProtein, int i, int c, int o, int j)
{
  checkBond(pProtein, i, c, o, j, N, H);

  switch (pProtein->resKey(j))
  {
  case SER:
  case THR:
    checkBond(pProtein, i, c, o, j, OG1, HG11);
    break;
  case ARG:
    checkBond(pProtein, i, c, o, j, NE2, HE11);
    checkBond(pProtein, i, c, o, j, NH1, HH11);
    checkBond(pProtein, i, c, o, j, NH1, HH12);
    checkBond(pProtein, i, c, o, j, NH2, HH21);
    checkBond(pProtein, i, c, o, j, NH2, HH21);
    break;
  case LYS:
    checkBond(pProtein, i, c, o, j, NZ1, HZ11);
    checkBond(pProtein, i, c, o, j, NZ1, HZ12);
    checkBond(pProtein, i, c, o, j, NZ1, HZ13);
    break;
  case ASN:
    checkBond(pProtein, i, c, o, j, ND2, HD21);
    checkBond(pProtein, i, c, o, j, ND2, HD22);
    break;
  case GLN:
    checkBond(pProtein, i, c, o, j, NE2, HE21);
    checkBond(pProtein, i, c, o, j, NE2, HE22);
    break;
  case HIS:
    checkBond(pProtein, i, c, o, j, ND1, HD11);
    checkBond(pProtein, i, c, o, j, NE2, HE21);
    break;
  case TRP:
    checkBond(pProtein, i, c, o, j, NE1, HE11);
    break;
  case TYR:
    checkBond(pProtein, i, c, o, j, OH1, HH11);
    break;
  }
}



void HBondList::countBetweenResidues(Protein* pProtein, int i, int j)
{
  countDonors(pProtein, i, C, O, j);

  switch (pProtein->resKey(i))
  {
  case SER:
  case THR:
    countDonors(pProtein, i, CB, OG1, j);
    break;
  case ASP:
    countDonors(pProtein, i, CG1, OD1, j);
    countDonors(pProtein, i, CG1, OD2, j);
    break;
  case ASN:
    countDonors(pProtein, i, CG1, OD1, j);
    break;
  case GLU:
    countDonors(pProtein, i, CD1, OE1, j);
    countDonors(pProtein, i, CD1, OE2, j);
    break;
  case GLN:
    countDonors(pProtein, i, CD1, OE1, j);
    break;
  case TYR:
    countDonors(pProtein, i, CZ1, OH1, j);
    break;
  }
}




void HBondList::countBetweenBackbones(Protein* pProtein, int res1, int res2)
{
  checkBond(pProtein, res1, C, O, res2, N, H);
}



void HBondList::countWithinResidue(Protein* pProtein, int i)
{
  switch (pProtein->resKey(i))
  {
  case SER:
  case THR:
    checkBond(pProtein, i, C, O, i, OG1, HG11);
    break;
  case ASP:
    checkBond(pProtein, i, CG1, OD1, i, N, H);
    checkBond(pProtein, i, CG1, OD2, i, N, H);
    break;
  case ASN:
    checkBond(pProtein, i, CG1, OD1, i, N, H);
    checkBond(pProtein, i, C, O, i, ND2, HD21);
    checkBond(pProtein, i, C, O, i, ND2, HD22);
    break;
  case GLU:
    checkBond(pProtein, i, CD1, OE1, i, N, H);
    checkBond(pProtein, i, CD1, OE2, i, N, H);
    break;
  case GLN:
    checkBond(pProtein, i, CD1, OE1, i, N, H);
    checkBond(pProtein, i, C, O, i, NE2, HE21);
    checkBond(pProtein, i, C, O, i, NE2, HE22);
    break;
  case TYR:
    checkBond(pProtein, i, CZ1, OH1, i, N, H);
    checkBond(pProtein, i, C, O, i, OH1, HH11);
    break;
    break;
  case ARG:
    checkBond(pProtein, i, C, O, i, NE2, HE11);
    checkBond(pProtein, i, C, O, i, NH1, HH11);
    checkBond(pProtein, i, C, O, i, NH1, HH12);
    checkBond(pProtein, i, C, O, i, NH2, HH21);
    checkBond(pProtein, i, C, O, i, NH2, HH21);
    break;
  case LYS:
    checkBond(pProtein, i, C, O, i, NZ1, HZ11);
    checkBond(pProtein, i, C, O, i, NZ1, HZ12);
    checkBond(pProtein, i, C, O, i, NZ1, HZ13);
    break;
  case HIS:
    checkBond(pProtein, i, C, O, i, ND1, HD11);
    checkBond(pProtein, i, C, O, i, NE2, HE21);
    break;
  case TRP:
    checkBond(pProtein, i, C, O, i, NE1, HE11);
    break;
  }
}



// weak HBondList



void WeakHBondList::countBond(Protein* pProtein, int i, int c, int o, int j, int n, int h)
{

  if (pos_dist_squared(pProtein->atom(j, CA).pos, pProtein->atom(i, CA).pos) > CA_DISTANCE_SQ)
    return;

  if (!pProtein->hasAtom(i, c) || !pProtein->hasAtom(i, o)
   || !pProtein->hasAtom(j, n) || !pProtein->hasAtom(j, h))
    return;

  double ho = pos_distance(pProtein->atom(j, h).pos, pProtein->atom(i, o).pos);
  double hc = pos_distance(pProtein->atom(j, h).pos, pProtein->atom(i, c).pos);
  double no = pos_distance(pProtein->atom(j, n).pos, pProtein->atom(i, o).pos);
  double nc = pos_distance(pProtein->atom(j, n).pos, pProtein->atom(i, c).pos);

  // make sure it is a hydrogen bond, i.e. the h is between the donors
  if (ho > nc) return;

  // the kabsch-sanders h-bond function!!!!
  double energy = (long)floor(Q_SQ/ho - Q_SQ/hc + Q_SQ/nc - Q_SQ/no + 0.5);

  if (energy > WEAK_HBOND_ENERGY_HIGH) return;

  // add this h-bond to the list
  BondStruct bond(i,c,o,j,n,h);
  list.push_back(bond);
}



void WeakHBondList::countDonors(Protein* pProtein, int i, int c, int o, int j)
{
  countBond(pProtein, i, c, o, j, CA, HA1);

  if (pProtein->resKey(j) == PRO) {
    countBond(pProtein, i, c, o, j, CD1, HD11);
    countBond(pProtein, i, c, o, j, CD1, HD12);
  } 
  if (pProtein->resKey(j) == GLY)
    countBond(pProtein, i, c, o, j, CA, HA2);

}


void WeakHBondList::countBetweenResidues(Protein* pProtein, int i, int j)
{
  countDonors(pProtein, i, C, O, j);

  switch (pProtein->resKey(i))
  {
  case SER:
  case THR:
    countDonors(pProtein, i, CB, OG1, j);
    break;
  case ASP:
    countDonors(pProtein, i, CG1, OD1, j);
    countDonors(pProtein, i, CG1, OD2, j);
    break;
  case ASN:
    countDonors(pProtein, i, CG1, OD1, j);
    break;
  case GLU:
    countDonors(pProtein, i, CD1, OE1, j);
    countDonors(pProtein, i, CD1, OE2, j);
    break;
  case GLN:
    countDonors(pProtein, i, CD1, OE1, j);
    break;
  case TYR:
    countDonors(pProtein, i, CZ1, OH1, j);
    break;
  }
}



void WeakHBondList::countBetweenBackbones(Protein* pProtein, int res1, int res2)
{
  countBond(pProtein, res1, C, O, res2, CA, HA1);
  if (pProtein->resKey(res2) == PRO)
    countBond(pProtein, res1, C, O, res2, CD1, HD11);
  if (pProtein->resKey(res2) == GLY)
    countBond(pProtein, res1, C, O, res2, CA, HA2);
}



// weak HBondList




void DisulfideList::countBetweenResidues(Protein* pProtein, int i, int j)
{
  if (pProtein->resKey(i) != CYS) return;
  if (pProtein->resKey(j) != CYS) return;

  if (pos_dist_squared(pProtein->atom(j, SG1).pos, pProtein->atom(i, SG1).pos)  
     > DISULFIDE_DISTANCE_SQ)
    return;

  // add this h-bond to the list
  BondStruct bond(i, CB, SG1,j, CB, SG1);
  list.push_back(bond);
}

