//
//  ***************************************************************************
//
//  display.cpp
//  (C) 2004 Bosco K. Ho 
//
//  ProteinDisplay is a widget that displays a pProtein in an allegro
//  bitmap. ProteinDisplay uses GraphicsView to convert 3d to 2d
//  drawing primitives; handles mouse input to rotate, zoom, and
//  to choose different active residues by clicking; and 
//  provides various options to display the pProtein.
//
//  ***************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//


#include "display.h"

using namespace std;



bool is_in_sphere(Vector3d& v1, Vector3d& v2)
{
  return(pos_dist_squared(v1, v2) < CA_DIST_SQ);
}



int atom_color(int resKey, int atomKey)
{
  switch (getAtomName(resKey, atomKey)[1])
  {
    case 'C':
      return LIGHT_GREY;
      break;
    case 'H':
      return  WHITE;
      break;
    case 'O':
      return  RED;
      break;
    case 'N':
      return  BLUE;
      break;
    case 'S':
      return  YELLOW;
      break;
  }
  return LIGHT_GREY;
}

               
               
ProteinDisplay::ProteinDisplay(Protein* p, int x, int y, int w, int h)
   : Widget(x, y, w, h), pProtein(p), isCentered(false), 
   stereoState(STEREO_NONE), isHbond(true), isSidechain(true),
   isSolid(true), isPeptide(true), isHydrogen(true), pickState(PICK_RESIDUE),
   showState(SHOW_ALL), mMeasureMarker(), mNLoopSolution(0),
   anchorRes1(-1), anchorRes2(-1), anchorRes3(-1)
{
  mGraphicsView.setStereoSeparation(SCREEN_W / 3);
  initializeFocus();
}



void ProteinDisplay::countBondsBetweenRes(int i, int j)
{
  if (isSidechain)
  {
    mHBondList.countBetweenResidues(pProtein, i, j);
    mWeakHBondList.countBetweenResidues(pProtein, i, j);
  }
  else
  {
    mHBondList.countBetweenBackbones(pProtein, i, j);
    mWeakHBondList.countBetweenBackbones(pProtein, i, j);
  }
}



void ProteinDisplay::countBonds()
{
  mHBondList.clear();
  mWeakHBondList.clear();
  mDisulfideList.clear();

  if (!isHbond) return;

  if (showState == SHOW_SPHERE)
  {
    for (int i = 0; i < pProtein->nRes(); i++)
    {
      Vector3d& focusAtomPos = pProtein->atom(focusRes, focusAtom).pos;
      if (resInSphere(i, focusAtomPos))
        if (isSidechain)
          mHBondList.countWithinResidue(pProtein, i);
      for (int j=i+1; j< pProtein->nRes(); j+=1)
        if (resInSphere(j, focusAtomPos))
        {
          mDisulfideList.countBetweenResidues(pProtein, i, j);
          countBondsBetweenRes(j, i);
          if (j > i+1)
            countBondsBetweenRes(i, j);
        }
    }
  }
  else
  {
    int show_left_res  = 0;
    int show_right_res = pProtein->nRes() - 1;

    if (showState == SHOW_SEGMENT)
    {
      show_left_res = viewLeftRes;
      show_right_res = viewRightRes;
    }

    for (int i = 0; i < pProtein->nRes(); i++)
    {
      if ( (i >= show_left_res) && (i <= show_right_res) )
        if (isSidechain)
          mHBondList.countWithinResidue(pProtein, i);

      for (int j=i+1; j < pProtein->nRes(); j+=1)
      if ( (j >= show_left_res) && (j <= show_right_res) )
        {
          mDisulfideList.countBetweenResidues(pProtein, i, j);
          countBondsBetweenRes(j, i);
          if (j > i+1)
            countBondsBetweenRes(i, j);
        }
    }
  }
}



// sets the focus to the residue closest to the center of mass
void ProteinDisplay::initializeFocus(void)
{
  if (pProtein->isEmpty()) 
    return;
  pProtein->computeCenterOfMass();

  activeLeftRes = 0;

  Vector3d centerOfMass = pProtein->getCenterOfMass();
  if (!pProtein->isEmpty())
  {
    double least = pos_dist_squared(centerOfMass, pProtein->atom(0, CA).pos);
    activeLeftRes = 0;
    for (int i = 0; i < pProtein->nRes(); i++)
    {
      double new_least = pos_dist_squared(centerOfMass, pProtein->atom(i, CA).pos);
      if (new_least < least)
      {
        activeLeftRes = i;
        least = new_least;
      }
    }
  }

  activeRightRes  = activeLeftRes;
  focusRes        = activeLeftRes;
  focusAtom       = CA;

  if (SCROLLBAR_SIZE < pProtein->nRes())
  {
    viewLeftRes = activeLeftRes - SCROLLBAR_SIZE/2;
    if (viewLeftRes < 0) 
      viewLeftRes = 0;
    viewRightRes = viewLeftRes + SCROLLBAR_SIZE - 1;
    while (viewRightRes >= pProtein->nRes())
    {
      viewLeftRes--;
      viewRightRes--;
    }
  }
  else
  {
    viewRightRes = pProtein->nRes() - 1;
    viewLeftRes = 0;
  }
  
  resetFocus();

  mCameraLook = mFocusAxis.scaled(-pProtein->maxLength() * 2.0);
  mCameraUp = mFocusUp.scaled(1.0);

  computeCameraFromFocus();
}



// computes a new transformation matrix based on the given
// focus position and camera-look and camera-up vectors
void ProteinDisplay::computeCameraFromFocus(void)
{
  Vector3d mCameraPos = mFocusPos - mCameraLook;
  mGraphicsView.isStereo = stereoState != STEREO_NONE;
  if (stereoState == STEREO_CROSSEYE) 
    mGraphicsView.setStereoViewAngle(8);
  if (stereoState == STEREO_WALLEYE) 
    mGraphicsView.setStereoViewAngle(-8);
  mGraphicsView.calculateMatrix(mCameraPos, mCameraLook, mCameraUp);
}



// given the focusRes and focusAtom values, changes the focusPos
// to that focusAtom. but if the option is to center then calculates
// the current center of mass based on user selected options
// once the focus is calculated, does house-keeping, calculates the
// new camera position and counts H-bonds
void ProteinDisplay::resetFocus(void)
{
  if (pProtein->isEmpty()) return;

  Vector3d ca_n = pProtein->atom(focusRes, N).pos - pProtein->atom(focusRes, CA).pos;
  Vector3d ca_c = pProtein->atom(focusRes, C).pos - pProtein->atom(focusRes, CA).pos;
  mFocusUp      = -(ca_n + ca_c);
  mFocusAxis    = cross(ca_n, ca_c);

  if (!isCentered)
    mFocusPos = pProtein->atom(focusRes, focusAtom).pos;
  else
  {
    if (showState == SHOW_SPHERE)
    {
      Vector3d& focusAtomPos = pProtein->atom(focusRes, focusAtom).pos;
      Vector3d mCenterOfMass(0.0, 0.0, 0.0);
      double nAtom = 0;

      for (int i = 0; i < pProtein->nRes(); i++)
        if (resInSphere(i, focusAtomPos))
        {
          AtomIter iter;
          for (iter = pProtein->resAtomBegin(i); iter != pProtein->resAtomEnd(i); iter++)
          {
            mCenterOfMass += iter->pos;
            nAtom += 1.0;
          }

        }
      mCenterOfMass /= nAtom;

      mFocusPos = mCenterOfMass;
    }
    else
    {
      int show_left_res;
      int show_right_res;

      if (showState == SHOW_ALL)
      {
        show_left_res  = 0;
        show_right_res = pProtein->nRes() - 1;
      }
      if (showState == SHOW_SEGMENT)
      {
        show_left_res  = viewLeftRes;
        show_right_res = viewRightRes;
      }

      Vector3d mCenterOfMass(0.0, 0.0, 0.0);
      double nAtom = 0;

      for (int i = show_left_res; i <= show_right_res; i++)
      {
        AtomIter iter;
        for (iter = pProtein->resAtomBegin(i); iter != pProtein->resAtomEnd(i); iter++)
        {
          mCenterOfMass += iter->pos;
          nAtom += 1.0;
        }
      }
      mCenterOfMass /= (nAtom);
      mFocusPos = mCenterOfMass;
    }
  }

  computeCameraFromFocus();
  countBonds();
}



// since we have a way of defining a standard reference frame w.r.t to each amino acid
// backbone, this function jumps to another focus_position but tries to
// maintain the relative orientation of the camera to the standard reference frame
void ProteinDisplay::resetFocusWithSameOrientation(void)
{
  /* get relative angles of old focus */
  double camera_look_a = vec_angle(mFocusUp, mCameraLook);
  double camera_look_b = vec_dihedral(mCameraLook, mFocusUp, mFocusAxis);

  double zoom   = mCameraLook.length();
  double up_a   = vec_dihedral(mCameraUp, mCameraLook, mFocusUp);

  resetFocus();

  /* reconstruct mCameraLook and mCameraUp using the relative angles of old focus */
  mCameraLook = mFocusAxis.rotated(camera_look_b, mFocusUp);

  Vector3d perp = cross(mFocusUp, mCameraLook);
  mCameraLook.rotate(-(90 - camera_look_a), perp);
  mCameraLook.scaleTo(zoom);

  perp = cross(mFocusUp, mCameraLook);
  mCameraUp = mCameraLook.rotated(-90, perp);
  mCameraUp.rotate(up_a, mCameraLook);
  computeCameraFromFocus();
}



int ProteinDisplay::handle(int event)
{
  if (pProtein->isEmpty())
    return 0;
    
  switch (event)
  {
    case WGT_MOUSE_B1_RELEASE:
      {
        if ( (mHighlightRes != -1) 
          && (mSavePickRes == mHighlightRes) 
          && (mSavePickAtom == mHighlightAtom) )
        {
          if (pickState == PICK_DISTANCE)
          {
            AtomMarker highlight;
            highlight.res  = mHighlightRes;
            highlight.atom = mHighlightAtom;
   
            bool isDelete = false;
            vector<AtomMarker>::iterator i;
            for (i = mMeasureMarker.begin(); i != mMeasureMarker.end(); i++)
            {
              if ( (i->res == highlight.res) && (i->atom == highlight.atom) )
              {
                mMeasureMarker.erase(i);
                isDelete = true;
                break;
              }
            }
            
            if (!isDelete)
            {
              int n = mMeasureMarker.size();
              if (n == 4)
              {
                for (int i = 0; i < 3; i++)
                  mMeasureMarker[i] = mMeasureMarker[i+1];
                mMeasureMarker.pop_back();
              }
           
              mMeasureMarker.push_back(highlight);
            }
          }
          else if (pickState == PICK_ANCHOR)
          {
            if (mHighlightRes >= 0)
              if ( (pProtein->resKey(mHighlightRes) != ACE)
                && (pProtein->resKey(mHighlightRes) != NME)
                && (pProtein->resKey(mHighlightRes) != PRO))
              {  
                if (mHighlightRes == anchorRes3)
                {
                  anchorRes3 = -1;
                  mNLoopSolution = 0;
                }
                else if (mHighlightRes == anchorRes2)
                {
                  anchorRes2 = anchorRes3;
                  anchorRes3 = -1;
                  mNLoopSolution = 0;
                }
                else if (mHighlightRes == anchorRes1)
                {
                  anchorRes1 = anchorRes2;
                  anchorRes2 = anchorRes3;
                  anchorRes3 = -1;
                  mNLoopSolution = 0;
                }
                else if (anchorRes1 == -1)
                {
                  anchorRes1 = mHighlightRes;
                }
                else if (anchorRes2 == -1)
                {
                  if (mHighlightRes > anchorRes1)
                    anchorRes2 = mHighlightRes;
                  else
                  {
                    anchorRes2 = anchorRes1;
                    anchorRes1 = mHighlightRes;
                  }
                }
                else if (anchorRes3 == -1)
                {
                  if (mHighlightRes < anchorRes1)
                  {
                    anchorRes3 = anchorRes2;
                    anchorRes2 = anchorRes1;
                    anchorRes1 = mHighlightRes;
                  }
                  else if (mHighlightRes < anchorRes2)
                  {
                    anchorRes3 = anchorRes2;
                    anchorRes2 = mHighlightRes;
                  }
                  else
                  {
                    anchorRes3 = mHighlightRes;
                  }
                  setNewAnchor();
                }
                else
                {
                  if (mHighlightRes > anchorRes3)
                    anchorRes3 = mHighlightRes;
                  else if (mHighlightRes < anchorRes1)
                    anchorRes1 = mHighlightRes;
                  else
                    anchorRes2 = mHighlightRes;
                  setNewAnchor();
                }
            }

          }
          else
          {
              activeLeftRes  = mHighlightRes;
              activeRightRes = mHighlightRes;
          }
        }
      }
      break;

    case WGT_MOUSE_B1_DOWN:
      {
        mSaveMouseY = mouse_y;
        mSaveMouseX = mouse_x;
        
        if (mHighlightRes != -1)
        {
          mSavePickRes  = mHighlightRes;
          mSavePickAtom = mHighlightAtom;
        } 
      }
      break;
    
    case WGT_MOUSE_B1_DRAG:
      {
        double rotate = (mouse_x - mSaveMouseX) / 3.0;
        mCameraLook.rotate(-rotate, mCameraUp);
   
        Vector3d norm = cross(mCameraLook, mCameraUp);
   
        double pitch  = (mouse_y - mSaveMouseY) / 3.0;
        mCameraLook.rotate(-pitch, norm);
        mCameraUp.rotate(-pitch, norm);
   
        computeCameraFromFocus();
   
        mSaveMouseY = mouse_y;
        mSaveMouseX = mouse_x;
      }
      break;

    case WGT_MOUSE_B2_DOWN:
      {
        mSaveMouseX = mouse_x;
        mSaveMouseY = mouse_y;
        if (mHighlightRes != -1)
        {
          mSavePickRes  = mHighlightRes;
          mSavePickAtom = mHighlightAtom;
        }
      }
      break;

    case WGT_MOUSE_B2_RELEASE:
      {
        if (mHighlightRes != -1) 
          if ((mSavePickRes == mHighlightRes) && (mSavePickAtom == mHighlightAtom))
          {
            focusRes  = mHighlightRes;
            focusAtom = mHighlightAtom;
            resetFocus();
          }
      }
      break;
      
    case WGT_MOUSE_B2_DRAG:
      {
        int new_zoom = (int)(mCameraLook.length() + (mouse_y - mSaveMouseY));
        if (new_zoom > 0) 
          mCameraLook.scaleTo(new_zoom);
        mCameraUp.rotate((mouse_x - mSaveMouseX), mCameraLook);
        computeCameraFromFocus();
   
        mSaveMouseY = mouse_y;
        mSaveMouseX = mouse_x;
      }
      break;

    default:
      return 0;
  }
  return 1;
}



bool ProteinDisplay::resInSphere(int i, Vector3d& rFocus)
{
  AtomIter iter;
  if (pos_distance(pProtein->atom(i, CA).pos, rFocus) > CA_DIST * 2) 
    return false;
  for (iter = pProtein->resAtomBegin(i); iter != pProtein->resAtomEnd(i); iter++)
  {
    if (is_in_sphere(iter->pos, rFocus))
      return true;
  }
  return false;
}



void ProteinDisplay::projectProteinToGraphicsView()
{
  mGraphicsView.clear();

  // clear highlight check;
  mHighlightRes = -1;

  // draw the h-bonds
  for (unsigned int i = 0; i < mHBondList.list.size(); i++)
  {
    if (isHydrogen)
      projectLine(mHBondList.list[i].acceptorRes, mHBondList.list[i].acceptor,
                  mHBondList.list[i].donorRes, mHBondList.list[i].hydrogen, ORANGE);
    else
      projectLine(mHBondList.list[i].acceptorRes, mHBondList.list[i].acceptor,
                  mHBondList.list[i].donorRes, mHBondList.list[i].donor, ORANGE);
  }


  // draw the weak h-bonds
  for (unsigned int i = 0; i < mWeakHBondList.list.size(); i++)
  {
    if (isHydrogen)
      projectLine(mWeakHBondList.list[i].acceptorRes, mWeakHBondList.list[i].acceptor,
                  mWeakHBondList.list[i].donorRes, mWeakHBondList.list[i].hydrogen, LIGHT_ORANGE);
    else
      projectLine(mWeakHBondList.list[i].acceptorRes, mWeakHBondList.list[i].acceptor,
                  mWeakHBondList.list[i].donorRes, mWeakHBondList.list[i].donor, LIGHT_ORANGE);
  }

  
  // draw the disulfide bonds
  for (unsigned int i = 0; i < mDisulfideList.list.size(); i++)
  {
    projectLine(mDisulfideList.list[i].acceptorRes, mDisulfideList.list[i].acceptor,
                    mDisulfideList.list[i].donorRes, mDisulfideList.list[i].hydrogen, YELLOW);
  }

  
  // draw pProtein
  if (showState == SHOW_SPHERE)
  {
    Vector3d& focusAtomPos = pProtein->atom(focusRes, focusAtom).pos;
    if (isPeptide)
      for (int i = 1; i < pProtein->nRes(); i++)
        if ( resInSphere(i-1, focusAtomPos) || resInSphere(i, focusAtomPos) )
        {
          projectPeptidePlane(i);
        }

    for (int i = 0; i < pProtein->nRes() - 1; i++)
      if ( resInSphere(i, focusAtomPos) || resInSphere(i+1, focusAtomPos) )
      {
        if ( (i >= 0) && (i < pProtein->nRes() - 1) )
          projectBond(i, C, i+1, N);
      }

    for (int i = 0; i < pProtein->nRes(); i++)
      if ( resInSphere(i, focusAtomPos) )
      {
        projectResidue(i);
      }
  }
  else
  {
    int show_left_res;
    int show_right_res;

    if (showState == SHOW_ALL)
    {
      show_left_res  = 0;
      show_right_res = pProtein->nRes() - 1;
    }
    if (showState == SHOW_SEGMENT)
    {
      show_left_res  = viewLeftRes;
      show_right_res = viewRightRes;
    }

    if (isPeptide)
      for (int i = show_left_res; i <= show_right_res + 1; i++)
      {
        if ( (i > 0) && (i < pProtein->nRes()) ) 
          projectPeptidePlane(i);
      }

    for (int i = show_left_res - 1; i <= show_right_res; i++)
    {
      if ( (i >= 0) && (i + 1 < pProtein->nRes()) )
        projectBond(i, C, i + 1, N);
    }

    for (int i = show_left_res; i <= show_right_res; i++)
    {
      projectResidue(i);
    }
    if (show_left_res > 0)
      projectBlockedAtom(show_left_res - 1, C, PURPLE, 1);
    if (show_right_res < pProtein->nRes() - 1)
      projectBlockedAtom(show_right_res + 1, N, PURPLE, 1);
  }

  // draw highlights for active residues
  for (int i = activeLeftRes; i <= activeRightRes; i+=1)
    projectHighlightAtom(i, CA, GREEN, 2);

  projectAnchor();
  projectMeasure();

  // draw highlights for rotation center
  if (focusRes >= 0) 
  {
    projectHighlightAtom(focusRes, focusAtom, PURPLE, 3);
    Vector3d& pos = pProtein->atom(focusRes, focusAtom).pos;
    if (showState == SHOW_SPHERE)
    {
      mGraphicsView.projectEnlargedCircle(pos, CA_DIST, 1, PURPLE);
    }
    if (isCentered)
    {
      mGraphicsView.projectEnlargedCircle(mFocusPos, DISPLAY_ATOM_RADIUS, 1, PURPLE);
      mGraphicsView.projectThinLine(mFocusPos, pos, PURPLE);
   }
  }

  if (pickState == PICK_RESIDUE)
  {
    if (mHighlightRes >= 0)
      projectHighlightAtom(mHighlightRes, mHighlightAtom, PURPLE, 4);
  }
  else if (pickState == PICK_ANCHOR)
  {
    if (mHighlightRes >= 0)
      if ( (pProtein->resKey(mHighlightRes) != PRO)
        && (pProtein->resKey(mHighlightRes) != ACE)
        && (pProtein->resKey(mHighlightRes) != NME) )
        {
          projectHighlightAtom(mHighlightRes, CA, BLUE, 8);
          projectHighlightAtom(mHighlightRes, CA, BLUE, 9);
          mGraphicsView.projectThinLine(pProtein->atom(mHighlightRes, CA).pos,
                                        pProtein->atom(mHighlightRes, mHighlightAtom).pos,
                                        BLUE);
        }
  }
  

  if (mHighlightRes != -1)
  {
    mHighlightString = pProtein->resAtomName(mHighlightRes, mHighlightAtom);
    Vector3d& p = pProtein->atom(mHighlightRes, mHighlightAtom).pos;
    mGraphicsView.projectString(p, LIGHT_GREY, &mHighlightString);
  }

  mGraphicsView.sort();
}



void ProteinDisplay::draw(BITMAP *bmp)
{
  mHighlightRes = -1;
  if (pProtein->isEmpty()) 
    return;
  projectProteinToGraphicsView();
  mGraphicsView.draw(bmp);
  stringout(bmp, font, "orientated on " + 
    pProtein->resAtomName(focusRes, focusAtom), 
    15, SCREEN_H - 12, PURPLE);
}



//  Translating protein parts into 3d graphics primitives



void ProteinDisplay::projectLine(int i, int a, int j, int b, int color)
{
  if (pProtein->hasAtom(i,a) && pProtein->hasAtom(j,b))
  {
    Vector3d& p1 = pProtein->atom(i,a).pos;
    Vector3d& p2 = pProtein->atom(j,b).pos;
    if (!isSolid)
      mGraphicsView.projectThinLine(p1, p2, color);
    else
      mGraphicsView.projectThickLine(p1, p2, color, DISPLAY_LINE_WIDTH);
  }
}



void ProteinDisplay::projectBond(int i, int a, int j, int b)
{
  if ( !isHydrogen )
    if ( (getAtomName(pProtein->resKey(i), pProtein->atom(i,a).key)[1] == 'H' ) ||
         (getAtomName(pProtein->resKey(j), pProtein->atom(j,b).key)[1] == 'H' ) )
      return;

  if (pProtein->hasAtom(i,a) && pProtein->hasAtom(j,b))
  {
    Vector3d& p1 = pProtein->atom(i,a).pos;
    Vector3d& p2 = pProtein->atom(j,b).pos;

    if (!isSolid)
    {
      Vector3d mid = (p1 + p2) * 0.5;
      int color = atom_color(pProtein->resKey(i), a);
      if ( (i >= activeLeftRes) && (i <= activeRightRes) && (color == LIGHT_GREY) ) 
        mGraphicsView.projectThinLine(p1, mid, GREEN);
      else
        mGraphicsView.projectThinLine(p1, mid, atom_color(pProtein->resKey(i), a));

      color = atom_color(pProtein->resKey(j), b);
      if ( (j >= activeLeftRes) && (j <= activeRightRes) && (color == LIGHT_GREY) ) 
        mGraphicsView.projectThinLine(p2, mid, GREEN);
      else
        mGraphicsView.projectThinLine(p2, mid, atom_color(pProtein->resKey(j), b));
    }
    else
    {
      if ( ((i >= activeLeftRes) && (i <= activeRightRes))
        || ((j >= activeLeftRes) && (j <= activeRightRes)) )
        mGraphicsView.projectThickLine(p1, p2, GREEN, DISPLAY_LINE_WIDTH);
      else
        mGraphicsView.projectThickLine(p1, p2, GREY, DISPLAY_LINE_WIDTH);
    }
  }
}



void ProteinDisplay::projectAtom(int r, int a)
{
  Vector3d& p = pProtein->atom(r, a).pos;

  if (isSolid)
     mGraphicsView.projectFilledCircle(p, DISPLAY_ATOM_RADIUS, atom_color(pProtein->resKey(r), a));

  int zOut;
  if (mGraphicsView.isInCircle(p, DISPLAY_ATOM_RADIUS, mouse_x, mouse_y, &zOut))
  {
    if ( (mHighlightRes == -1) || (zOut < mHighlightZ) )
    {
      mHighlightRes  = r;
      mHighlightAtom = a;
      mHighlightZ    = zOut;
    }
  }
}



void ProteinDisplay::projectHighlightAtom(int res, int a, int color, int highlight_size)
{
  mGraphicsView.projectEnlargedCircle(pProtein->atom(res, a).pos, DISPLAY_ATOM_RADIUS, highlight_size, color);
}



void ProteinDisplay::projectBlockedAtom(int res, int a, int color, int highlight_size)
{
  mGraphicsView.projectEnlargedFilledCircle(pProtein->atom(res, a).pos,
     DISPLAY_ATOM_RADIUS, highlight_size, color);
}



void ProteinDisplay::projectPeptidePlane(int i)
{
  Vector3d v[4];

  if (pProtein->resKey(i) == PRO)
  {
    if ( pProtein->hasAtom(i-1, CA) 
      && pProtein->hasAtom(i-1,  O) 
      && pProtein->hasAtom(i,   CA) 
      && pProtein->hasAtom(i,  CD1) )
    {
      v[0] = pProtein->atom(i-1,  CA).pos;
      v[1] = pProtein->atom(i-1,   O).pos;
      v[2] = pProtein->atom(i,    CA).pos;
      v[3] = pProtein->atom(i,   CD1).pos;
      if (isSolid)
        mGraphicsView.projectQuad(v, DARK_GREY);
      else
      {
        mGraphicsView.projectWireframeQuad(v, DARK_GREY);
      }
    }
  }
  else
  {
    if ( pProtein->hasAtom(i-1, CA) 
      && pProtein->hasAtom(i-1,  O) 
      && pProtein->hasAtom(i,   CA) 
      && pProtein->hasAtom(i,    N) )
    {
      v[0] = pProtein->atom(i-1, CA).pos;
      v[1] = pProtein->atom(i-1,  O).pos;
      v[2] = pProtein->atom(i,   CA).pos;

      if (isHydrogen && pProtein->hasAtom(i, H))
        v[3] = pProtein->atom(i, H).pos;
      else
        v[3] = pProtein->atom(i, N).pos;
      if (isSolid)
        mGraphicsView.projectQuad(v, DARK_GREY);
      else
      {
        mGraphicsView.projectWireframeQuad(v, DARK_GREY);
      }
    }
  }
}



void ProteinDisplay::projectMeasure()
{
  int n = mMeasureMarker.size();
  for (int i = 0; i < n-1; i++)
    projectLine(mMeasureMarker[i].res, mMeasureMarker[i].atom,
                mMeasureMarker[i+1].res, mMeasureMarker[i+1].atom, PURPLE);
  for (int i = 0; i < n; i++) 
    projectBlockedAtom(mMeasureMarker[i].res, mMeasureMarker[i].atom, PURPLE, 0);
  if (mHighlightRes > -1)
    projectHighlightAtom(mHighlightRes, mHighlightAtom, RED, 4);

  if (0 == n) return;

  projectHighlightAtom(mMeasureMarker[n-1].res, mMeasureMarker[n-1].atom, RED, 4);

  if (n >= 2)
  {
    Vector3d& p0 = pProtein->atom(mMeasureMarker[0].res, mMeasureMarker[0].atom).pos; 
    Vector3d& p1 = pProtein->atom(mMeasureMarker[1].res, mMeasureMarker[1].atom).pos; 
    double dist = pos_distance(p0, p1);
    Vector3d pos = (p0 + p1) * 0.5;
    mMeasureDistString = double_to_string(dist, 0, 2);
    mGraphicsView.projectCentreString(pos, LIGHT_GREY, &mMeasureDistString);
  }
  if (n >= 3)
  {
    Vector3d& p0 = pProtein->atom(mMeasureMarker[0].res, mMeasureMarker[0].atom).pos; 
    Vector3d& p1 = pProtein->atom(mMeasureMarker[1].res, mMeasureMarker[1].atom).pos; 
    Vector3d& p2 = pProtein->atom(mMeasureMarker[2].res, mMeasureMarker[2].atom).pos; 
    double ang = pos_angle(p0, p1, p2);
    mMeasureAngleString = double_to_string(ang, 0, 1);
    mGraphicsView.projectCentreString(p1, LIGHT_GREY, &mMeasureAngleString);

    if (n == 3)
    {
      double dist = pos_distance(p1, p2);
      Vector3d pos = (p1 + p2) * 0.5;
      mMeasureDist2String = double_to_string(dist, 0, 2);
      mGraphicsView.projectCentreString(pos, LIGHT_GREY, &mMeasureDist2String);
    }
  }
  if (n >= 4)
  {
    Vector3d& p0 = pProtein->atom(mMeasureMarker[0].res, mMeasureMarker[0].atom).pos; 
    Vector3d& p1 = pProtein->atom(mMeasureMarker[1].res, mMeasureMarker[1].atom).pos; 
    Vector3d& p2 = pProtein->atom(mMeasureMarker[2].res, mMeasureMarker[2].atom).pos; 
    Vector3d& p3 = pProtein->atom(mMeasureMarker[3].res, mMeasureMarker[3].atom).pos; 
    double dihed = pos_dihedral(p0, p1, p2, p3);
    mMeasureDihedString = "(" + double_to_string(dihed, 0, 1) + ")";
    Vector3d pos = (p1 + p2) * 0.5;
    mGraphicsView.projectCentreString(pos, LIGHT_GREY, &mMeasureDihedString);
    
    double ang = pos_angle(p1, p2, p3);
    mMeasureAngle2String = double_to_string(ang, 0, 1);
    mGraphicsView.projectCentreString(p2, LIGHT_GREY, &mMeasureAngle2String);
    
    double dist = pos_distance(p2, p3);
    pos = (p2 + p3) * 0.5;
    mMeasureDist2String = double_to_string(dist, 0, 2);
    mGraphicsView.projectCentreString(pos, LIGHT_GREY, &mMeasureDist2String);
  }
}




void ProteinDisplay::projectAnchor()
{
  if (anchorRes1 >= 0)
  {
    projectBlockedAtom(anchorRes1, CA, BLUE, 0);
    projectHighlightAtom(anchorRes1, CA, BLUE, 4);
    projectHighlightAtom(anchorRes1, CA, BLUE, 8);
    projectHighlightAtom(anchorRes1, CA, BLUE, 9);
  }
  
  if (anchorRes2 >= 0)
  {
    projectLine(anchorRes1, CA, anchorRes2, CA, BLUE);
    projectBlockedAtom(anchorRes2, CA, BLUE, 0);
    projectHighlightAtom(anchorRes2, CA, BLUE, 4);
    projectHighlightAtom(anchorRes2, CA, BLUE, 8);
    projectHighlightAtom(anchorRes2, CA, BLUE, 9);
  }
  
  if (anchorRes3 >= 0)
  {
    projectLine(anchorRes2, CA, anchorRes3, CA, BLUE);
    projectBlockedAtom(anchorRes3, CA, BLUE, 0);
    projectHighlightAtom(anchorRes3, CA, BLUE, 4);
    projectHighlightAtom(anchorRes3, CA, BLUE, 8);
    projectHighlightAtom(anchorRes3, CA, BLUE, 9);
  }
  
}



void ProteinDisplay::projectResidue(int i)
{
  projectBonds(i);
  projectAtoms(i);
}



void ProteinDisplay::projectBonds(int i)
{
  projectBond(i, CA, i, HA1);
  projectBond(i, CA, i,   C);
  projectBond(i,  C, i,   O);
  projectBond(i, CA, i,   N);
  projectBond(i,  N, i,   H);

  if (pProtein->resKey(i) != GLY)
    projectBond(i, CA, i, CB);
  else
    projectBond(i, CA, i, HA2);

  // terminal chicanery
  if (pProtein->resKey(i) == ACE)
  {
    projectBond(i, CA, i, H1);
    projectBond(i, CA, i, H2);
    projectBond(i, CA, i, H3);
  }
  else if (pProtein->resKey(i) == NME)
  {
    projectBond(i, CA, i, H1);
    projectBond(i, CA, i, H2);
    projectBond(i, CA, i, H3);
  }
  else
  {
    projectBond(i, C,  i, OXT);
    projectBond(i, OXT,i, HXT);
    projectBond(i, N,  i, H2);
    projectBond(i, N,  i, H3);
  }

  if (pProtein->resKey(i) == PRO)
  {
    projectBond(i, CB,  i, CG1);
    projectBond(i, CG1, i, CD1);
    projectBond(i, CA,  i, N);
    projectBond(i, N,   i, CD1);

    projectBond(i, CD1, i, N);
    projectBond(i, CD1, i, HD11);
    projectBond(i, CD1, i, HD12);
    projectBond(i, CB,  i, HB1);
    projectBond(i, CB,  i, HB2);
    projectBond(i, CG1, i, HG11);
    projectBond(i, CG1, i, HG12);
  }

  if (pProtein->resKey(i) == CYS)
  {
    projectBond(i, CB, i, HB1);
    projectBond(i, CB, i, HB2);
    projectBond(i, CB, i, SG1);
    projectBond(i, SG1, i, HG11);
  }
  else if (isSidechain)
  {
    switch (pProtein->resKey(i))
    {
      case ALA:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);
        projectBond(i, CB, i, HB3);

        break;
      case SER:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);
        projectBond(i, CB,  i, OG1);
        projectBond(i, OG1, i, HG11);
        break;
      case ASP:
        projectBond(i, CB,  i, HB1);

        projectBond(i, CB,  i, HB2);
        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, OD1);
        projectBond(i, CG1, i, OD2);
        projectBond(i, OD1 ,i, HD11);
        projectBond(i, OD2 ,i, HD21);

        break;
      case ASN:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);
        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, OD1);
        projectBond(i, CG1, i, ND2);
        projectBond(i, ND2, i, HD21);
        projectBond(i, ND2, i, HD22);

        break;
      case GLU:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);
        projectBond(i, CB,  i, CG1);

        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);
        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, OE1);
        projectBond(i, CD1, i, OE2);
        projectBond(i, OE1 ,i, HE11);
        projectBond(i, OE2 ,i, HE21);

        break;
      case GLN:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);
        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);
        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, OE1);
        projectBond(i, CD1, i, NE2);
        projectBond(i, NE2, i, HE21);
        projectBond(i, NE2, i, HE22);

        break;
      case THR:

        projectBond(i, CB,  i, HB1);

        projectBond(i, CB,  i, OG1);
        projectBond(i, OG1, i, HG11);

        projectBond(i, CB,  i, CG2);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CG2, i, HG22);
        projectBond(i, CG2, i, HG23);

        break;
      case VAL:
        projectBond(i, CB,  i, HB1);

        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);
        projectBond(i, CG1, i, HG13);

        projectBond(i, CB,  i, CG2);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CG2, i, HG22);
        projectBond(i, CG2, i, HG23);

        break;
      case ILE:
        projectBond(i, CB,  i, HB1);

        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);


        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, HD11);
        projectBond(i, CD1, i, HD12);
        projectBond(i, CD1, i, HD13);

        projectBond(i, CB,  i, CG2);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CG2, i, HG22);
        projectBond(i, CG2, i, HG23);


        break;
      case LEU:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);

        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, HG11);

        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, HD11);
        projectBond(i, CD1, i, HD12);
        projectBond(i, CD1, i, HD13);
        projectBond(i, CG1, i, CD2);
        projectBond(i, CD2, i, HD21);
        projectBond(i, CD2, i, HD22);
        projectBond(i, CD2, i, HD23);

        break;
      case LYS:
        projectBond(i, CB,  i, HB1);
        projectBond(i, CB,  i, HB2);

        projectBond(i, CB,  i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);

        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, HD11);
        projectBond(i, CD1, i, HD12);

        projectBond(i, CD1, i, CE1);
        projectBond(i, CE1, i, HE11);
        projectBond(i, CE1, i, HE12);

        projectBond(i, CE1, i, NZ1);
        projectBond(i, NZ1, i, HZ11);
        projectBond(i, NZ1, i, HZ12);
        projectBond(i, NZ1, i, HZ13);

        break;
      case MET:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);

        projectBond(i, CG1, i, SD1);

        projectBond(i, SD1, i, CE1);
        projectBond(i, CE1, i, HE11);
        projectBond(i, CE1, i, HE12);
        projectBond(i, CE1, i, HE13);


        break;
      case ARG:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, HG11);
        projectBond(i, CG1, i, HG12);

        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, HD11);
        projectBond(i, CD1, i, HD12);

        projectBond(i, CD1, i, NE1);
        projectBond(i, NE1, i, HE11);

        projectBond(i, NE1, i, CZ1);

        projectBond(i, CZ1, i, NH1);
        projectBond(i, NH1, i, HH11);
        projectBond(i, NH1, i, HH12);

        projectBond(i, CZ1, i, NH2);
        projectBond(i, NH2, i, HH21);
        projectBond(i, NH2, i, HH22);
        break;
      case HIS:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, ND1);
        projectBond(i, ND1, i, CE1);
        projectBond(i, CE1, i, NE2);
        projectBond(i, NE2, i, CD2);
        projectBond(i, CD2, i, CG1);

        projectBond(i, ND1, i, HD11);
        projectBond(i, CE1, i, HE11);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CD2, i, HD21);
        projectBond(i, NE2, i, HE21);

        break;
      case TRP:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, NE1);
        projectBond(i, NE1, i, CE2);
        projectBond(i, CE2, i, CD2);
        projectBond(i, CD2, i, CG1);

        projectBond(i, CE2, i, CZ2);
        projectBond(i, CZ2, i, CH2);
        projectBond(i, CD2, i, CE3);
        projectBond(i, CE3, i, CZ3);
        projectBond(i, CZ3, i, CH2);

        projectBond(i, CD1, i, HD11);
        projectBond(i, NE1, i, HE11);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CE3, i, HE31);
        projectBond(i, CZ2, i, HZ21);
        projectBond(i, CZ3, i, HZ31);
        projectBond(i, CH2, i, HH21);
        break;
      case PHE:
        projectBond(i, CB, i, HB1);

        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, CE1);
        projectBond(i, CE1, i, CZ1);
        projectBond(i, CZ1, i, CE2);
        projectBond(i, CE2, i, CD2);
        projectBond(i, CD2, i, CG1);

        projectBond(i, CD1, i, HD11);
        projectBond(i, CE1, i, HE11);
        projectBond(i, CZ1, i, HZ11);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CD2, i, HD21);
        projectBond(i, CE2, i, HE21);

        break;
      case TYR:
        projectBond(i, CB, i, HB1);
        projectBond(i, CB, i, HB2);

        projectBond(i, CB, i, CG1);
        projectBond(i, CG1, i, CD1);
        projectBond(i, CD1, i, CE1);
        projectBond(i, CE1, i, CZ1);
        projectBond(i, CZ1, i, CE2);
        projectBond(i, CE2, i, CD2);
        projectBond(i, CD2, i, CG1);

        projectBond(i, CD1, i, HD11);
        projectBond(i, CE1, i, HE11);
        projectBond(i, CZ1, i, OH1);
        projectBond(i, OH1, i, HH11);
        projectBond(i, CG2, i, HG21);
        projectBond(i, CD2, i, HD21);
        projectBond(i, CE2, i, HE21);

        break;
    }
  }
}



void ProteinDisplay::projectAtoms(int i)
{
  for (AtomIter a = pProtein->resAtomBegin(i); a != pProtein->resAtomEnd(i); a++)
    if ( !(!isHydrogen && (getAtomName(a->resKey, a->key)[1] == 'H')) )
    {
      if ( (a->key > CB) && (a->resKey != PRO) && (a->resKey != CYS))
      {
        if (isSidechain) 
          projectAtom(i, a->key);
      }
      else
        projectAtom(i, a->key);
    }
}



void load_v_array(double v_array[], Vector3d& vec)
{
  v_array[0] = vec.x();
  v_array[1] = vec.y();
  v_array[2] = vec.z();
}



void load_vector3d(double v_array[], Vector3d& vec)
{
  vec.x(v_array[0]);
  vec.y(v_array[1]);
  vec.z(v_array[2]);
}



void extract_loop_closure_param(Protein& p, int res1, int res2, int res3)
{
  double b_len[6], b_ang[7], t_ang[2];

  // bond lengths
  b_len[0]= pos_distance(p.atom(res1, CA).pos, p.atom(res1, C).pos); 
  b_len[1]= pos_distance(p.atom(res1, C).pos,  p.atom(res2, N).pos); 
  b_len[2]= pos_distance(p.atom(res2, N).pos,  p.atom(res2, CA).pos); 
  b_len[3]= pos_distance(p.atom(res2, CA).pos, p.atom(res2, C).pos); 
  b_len[4]= pos_distance(p.atom(res2, C).pos,  p.atom(res3, N).pos); 
  b_len[5]= pos_distance(p.atom(res3, N).pos,  p.atom(res3, CA).pos); 

  // bond angles
  b_ang[0]= DEG2RAD*pos_angle(p.atom(res1, N).pos, p.atom(res1, CA).pos,
                              p.atom(res1, C).pos); // ang_nac;
  b_ang[1]= DEG2RAD*pos_angle(p.atom(res1, CA).pos, p.atom(res1, C).pos,  
                              p.atom(res2, N).pos); // ang_acn;
  b_ang[2]= DEG2RAD*pos_angle(p.atom(res1, C).pos,  p.atom(res2, N).pos,  
                              p.atom(res2, CA).pos); // ang_cna;
  b_ang[3]= DEG2RAD*pos_angle(p.atom(res2, N).pos,  p.atom(res2, CA).pos, 
                              p.atom(res2, C).pos); // ang_nac;
  b_ang[4]= DEG2RAD*pos_angle(p.atom(res2, CA).pos, p.atom(res2, C).pos,  
                              p.atom(res3, N).pos); // ang_acn;
  b_ang[5]= DEG2RAD*pos_angle(p.atom(res2, C).pos,  p.atom(res3, N).pos,  
                              p.atom(res3, CA).pos); // ang_cna;
  b_ang[6]= DEG2RAD*pos_angle(p.atom(res3, N).pos,  p.atom(res3, CA).pos, 
                              p.atom(res3, C).pos); // ang_nac;
   
  // torsion angles between anchor residues
  t_ang[0] = DEG2RAD*pos_dihedral(p.atom(res1, CA).pos,  p.atom(res1, C).pos, 
                                  p.atom(res2, N).pos,   p.atom(res2, CA).pos);
  t_ang[1] = DEG2RAD*pos_dihedral(p.atom(res2, CA).pos,  p.atom(res2, C).pos, 
                                  p.atom(res3, N).pos,   p.atom(res3, CA).pos);

  initialize_loop_closure(b_len, b_ang, t_ang);
}



/*
 * using the parameters extracted in extract_loop_closure_param,
 * calculates the solutions, which are placed in r_soln_n,
 * r_soln_a, r_soln_c 
 */
void ProteinDisplay::computeLoopClosureSolutions()
{
  double r_n1[3], r_a1[3], r_a3[3], r_c3[3];

  load_v_array(r_n1, pProtein->atom(anchorRes1, N).pos);
  load_v_array(r_a1, pProtein->atom(anchorRes1, CA).pos);

  load_v_array(r_a3, pProtein->atom(anchorRes3, CA).pos);
  load_v_array(r_c3, pProtein->atom(anchorRes3, C).pos);

  solve_3pep_poly(r_n1, r_a1, r_a3, r_c3, r_soln_n, r_soln_a, r_soln_c, &mNLoopSolution);

  if (mNLoopSolution == 0) 
    return;
  
  // work out which solution corresponds to current conformation
  mILoopSolution = 0;
  Vector3d ca;
  load_vector3d(r_soln_a[0][1], ca);
  double minDistDiff = pos_distance(pProtein->atom(anchorRes2,CA).pos, ca);
  for (int i=1; i<mNLoopSolution; i++)
  {
    load_vector3d(r_soln_a[i][1], ca);
    double distDiff = pos_distance(pProtein->atom(anchorRes2,CA).pos, ca);
    if (distDiff < minDistDiff)
    {
      minDistDiff = distDiff;
      mILoopSolution = i;
    }
  }
}



void ProteinDisplay::setNewAnchor()
{
  if (anchorRes3 < 0) 
    return;
  extract_loop_closure_param(*pProtein, anchorRes1, anchorRes2,
                             anchorRes3);
  computeLoopClosureSolutions();
}



void ProteinDisplay::changeAnchorPhiPsi(double phi, double psi)
{
  Protein& p = *pProtein;
  int res1 = anchorRes1;
  int res2 = anchorRes2;
  int res3 = anchorRes3;
  
  Protein temp = p;
  p.extractFragment(temp, res1, res3);
  for (int i=0; i<temp.nRes(); i++) 
  {
    int res = i + res1;
    if ( ((res > res1) && (res < res2)) || ((res > res2) && (res < res3)) )
      if ((res >=activeLeftRes) && (res <=activeRightRes))
      {
        if (temp.resKey(i) == PRO)
        {
          if (!temp.projectProlineRing(i,phi))
            return;
          temp.rotateProlinePhi(i, phi);
        }
        else
        {
          temp.rotatePhi(i, phi);
        }
        temp.rotatePsi(i, psi);
      }
  }
  extract_loop_closure_param(temp, 0, res2-res1, res3-res1);
  computeLoopClosureSolutions();
  
  if (mNLoopSolution <= 0)
  {
    setNewAnchor();
    return;
  }

  int k = mILoopSolution;
  
  Vector3d newC1, newN2, newCA2, newC2, newN3;

  load_vector3d(r_soln_c[k][0], newC1);
  load_vector3d(r_soln_n[k][1], newN2);
  load_vector3d(r_soln_a[k][1], newCA2);
  load_vector3d(r_soln_c[k][1], newC2);
  load_vector3d(r_soln_n[k][2], newN3);

  Vector3d axis, center;
  
  Vector3d oldC2 = p.atom(res2,C).pos;
  Vector3d oldN2 = p.atom(res2,N).pos;
  Vector3d oldCA2 = p.atom(res2,CA).pos;
    
  double dPhi1  = pos_dihedral(p.atom(res1,C).pos, p.atom(res1,N).pos, p.atom(res1,CA).pos, newC1);
  double dPsi3  = pos_dihedral(p.atom(res3,N).pos, p.atom(res3,CA).pos, p.atom(res3,C).pos, newN3);

  double newAlpha = pos_dihedral(p.atom(res1,N).pos, p.atom(res1,CA).pos, newC1, newN2);
  double newBeta = pos_dihedral(newC2, newN3, p.atom(res3,CA).pos, p.atom(res3,C).pos);

  // phi1 rotation
  center = p.atom(res1,CA).pos;
  axis = center - p.atom(res1,N).pos;
  Matrix3d transform1(axis, dPhi1, center);
  p.phi(res1) += dPhi1;
  p.atom(res1,C).pos.applyMatrix(transform1);
  p.atom(res1,O).pos.applyMatrix(transform1);
  p.atom(res1,HA1).pos.applyMatrix(transform1);
  for (AtomIter a = p.resAtomBegin(res1); a != p.resAtomEnd(res1); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(transform1);
  for (int i=res1+1; i<res2; i++)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(transform1);
  p.atom(res2,H).pos.applyMatrix(transform1);
  p.atom(res2,N).pos.applyMatrix(transform1);
  
  // intermediate phi/psi rotation between res1 and res2
  for (int i=res1+1; i<res2; i++)
  {
    if ((i >=activeLeftRes) && (i <=activeRightRes))
    { 
      // phi1 rotation
      center = p.atom(i,CA).pos;
      axis = center - p.atom(i,N).pos;
      if (p.resKey(i)== PRO)
        p.projectProlineRing(i, phi);
      Matrix3d transform2(axis, phi - p.phi(i), center);
      p.atom(i,C).pos.applyMatrix(transform2);
      p.atom(i,O).pos.applyMatrix(transform2);
      p.atom(i,HA1).pos.applyMatrix(transform2);
      for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
        if (a->key >= CB)
          a->pos.applyMatrix(transform2);
      for (int j=i+1; j<res2; j++)
      {
        for (AtomIter a = p.resAtomBegin(j); a != p.resAtomEnd(j); a++)
          a->pos.applyMatrix(transform2);
      }
      p.atom(res2,H).pos.applyMatrix(transform2);
      p.atom(res2,N).pos.applyMatrix(transform2);
      
      // psi1 rotation
      center = p.atom(i,C).pos;
      axis = center - p.atom(i,CA).pos;
      Matrix3d transform3(axis, psi-p.psi(i), center);
      p.psi(i) = psi;
      p.atom(i,O).pos.applyMatrix(transform3);
      for (int j=i+1; j<res2; j++)
        for (AtomIter a = p.resAtomBegin(j); a != p.resAtomEnd(j); a++)
          a->pos.applyMatrix(transform3);
      p.atom(res2,H).pos.applyMatrix(transform3);
      p.atom(res2,N).pos.applyMatrix(transform3);
    }
  }
  
  // psi1 rotation
  double oldAlpha = pos_dihedral(p.atom(res1,N).pos, p.atom(res1,CA).pos, 
                                 p.atom(res1,C).pos, p.atom(res2,N).pos);
  double dPsi1 = newAlpha - oldAlpha;
  double newPsi1 = dPsi1 + p.psi(res1);
  p.psi(res1) += dPsi1;
  center = p.atom(res1,C).pos;
  axis = center - p.atom(res1,CA).pos;
  Matrix3d transform4(axis, dPsi1, center);
  p.atom(res1,O).pos.applyMatrix(transform4);
  for (int i=res1+1; i<res2; i++)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(transform4);
  p.atom(res2,H).pos.applyMatrix(transform4);
  p.atom(res2,N).pos.applyMatrix(transform4);
  
  // rotate segment between res2 and res3
  
  // psi3 rotation
  center = p.atom(res3,CA).pos;
  axis = p.atom(res3,C).pos - center;
  Matrix3d transform5(axis, dPsi3, center);
  p.psi(res3) += dPsi3;
  p.atom(res3,N).pos.applyMatrix(transform5);
  p.atom(res3,H).pos.applyMatrix(transform5);
  p.atom(res3,HA1).pos.applyMatrix(transform5);
  for (AtomIter a = p.resAtomBegin(res3); a != p.resAtomEnd(res3); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(transform5);
  for (int i=res3-1; i>res2; i--)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(transform5);
  p.atom(res2,O).pos.applyMatrix(transform5);
  p.atom(res2,C).pos.applyMatrix(transform5);
 
  
  // intermediate phi/psi rotation between res3 and res2
  for (int i=res3-1; i>res2; i--)
  {
    if ((i >=activeLeftRes) && (i <=activeRightRes))
    { 
      // psi rotation
      center = p.atom(i,CA).pos;
      axis = p.atom(i,C).pos - center;
      double dpsi = (psi - p.psi(i));
      p.psi(i) = psi;
      Matrix3d transform6(axis, -dpsi, center);
      p.atom(i,N).pos.applyMatrix(transform6);
      p.atom(i,H).pos.applyMatrix(transform6);
      p.atom(i,HA1).pos.applyMatrix(transform6);
      for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
        if (a->key >= CB)
          a->pos.applyMatrix(transform6);
      for (int j=i-1; j>res2; j--)
        for (AtomIter a = p.resAtomBegin(j); a != p.resAtomEnd(j); a++)
          a->pos.applyMatrix(transform6);
      p.atom(res2,O).pos.applyMatrix(transform6);
      p.atom(res2,C).pos.applyMatrix(transform6);
      
      // phi rotation
      center = p.atom(i,N).pos;
      axis = p.atom(i,CA).pos - center;
      double dphi = phi - p.phi(i);
      if (p.resKey(i)== PRO)
        p.projectProlineRing(i, phi);
      Matrix3d transform7(axis, -dphi, center);
      p.atom(i,H).pos.applyMatrix(transform7);
      for (int j=i-1; j>res2; j--)
      {  
        for (AtomIter a = p.resAtomBegin(j); a != p.resAtomEnd(j); a++)
          a->pos.applyMatrix(transform7);
      }
      p.atom(res2,O).pos.applyMatrix(transform7);
      p.atom(res2,C).pos.applyMatrix(transform7);
    }
  }
  
  // phi3 rotation
  double oldBeta = pos_dihedral(p.atom(res2,C).pos,  p.atom(res3,N).pos, 
                                p.atom(res3,CA).pos, p.atom(res3,C).pos);
  double dPhi3 = newBeta - oldBeta;
  center = p.atom(res3,N).pos;
  axis = p.atom(res3,CA).pos - center;
  Matrix3d transform8(axis, -dPhi3, center);
  p.atom(res3,H).pos.applyMatrix(transform8);
  for (int i=res3-1; i>res2; i--)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(transform8);
  p.atom(res2,O).pos.applyMatrix(transform8);
  p.atom(res2,C).pos.applyMatrix(transform8);
  
  p.atom(res2,CA).pos = newCA2;
  

  // calculate orientation of res2 using oldC2 and oldN2
  center = p.atom(res2,CA).pos;

  Vector3d displacement = center - oldCA2;
  p.atom(res2,HA1).pos += displacement;
  for (AtomIter a = p.resAtomBegin(res2); a != p.resAtomEnd(res2); a++)
    if (a->key >= CB)
      a->pos += displacement;
  oldN2 += displacement;
  oldC2 += displacement;
    
  Vector3d old_ca_c = oldC2 - center;
  Vector3d new_ca_c = p.atom(res2,C).pos - center;

  axis = cross(old_ca_c, new_ca_c);
  double torsion4 = vec_dihedral(new_ca_c, axis, old_ca_c);
  
  Matrix3d transform10(axis, torsion4, center);  
  p.atom(res2,HA1).pos.applyMatrix(transform10);
  for (AtomIter a = p.resAtomBegin(res2); a != p.resAtomEnd(res2); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(transform10);
  oldN2.applyMatrix(transform10);

  Vector3d old_ca_n = oldN2 - center;
  Vector3d new_ca_n = p.atom(res2,N).pos - center;

  axis = new_ca_c;
  double torsion5 = vec_dihedral(new_ca_n, axis, old_ca_n);
  
  Matrix3d transform11(axis, torsion5, center);  
  p.atom(res2,HA1).pos.applyMatrix(transform11);
  for (AtomIter a = p.resAtomBegin(res2); a != p.resAtomEnd(res2); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(transform11);

  p.extractBackboneAngles();


  // rotate resdiues outside anchor 
  for (int i=activeLeftRes; i<=activeRightRes; i+=1)
    if ((i < res1) || (i > res3))
    {
      if (pProtein->resKey(i) == PRO)
        p.rotateProlinePhi(i, phi);
      else
        p.rotatePhi(i, phi);
      p.rotatePsi(i, psi);
    }
  
  setNewAnchor();
}




void ProteinDisplay::changeLoopSolution(int k)
{
  if (mNLoopSolution == 0) 
    return;
  if ((k < 0) || (k >= mNLoopSolution) || (k == mILoopSolution)) 
    return;
  
  mILoopSolution = k;
  
  Protein& p = *pProtein;
  
  int res1 = anchorRes1;
  int res2 = anchorRes2;
  int res3 = anchorRes3;
  
  Vector3d newC1, newCA2, newC2;

  load_vector3d(r_soln_c[k][0], newC1);
  load_vector3d(r_soln_a[k][1], newCA2);
  load_vector3d(r_soln_c[k][1], newC2);

  Vector3d center;
  Vector3d axis;
  
  Vector3d oldC1 = p.atom(res1,C).pos;
  Vector3d oldN3 = p.atom(res3,N).pos;

  // cary out first rotation between CA1 and CA3
  double torsion1 = pos_dihedral(p.atom(res2, CA).pos,  p.atom(res1, CA).pos, 
                                   p.atom(res3, CA).pos,  newCA2);
  center = p.atom(res1, CA).pos;
  axis = p.atom(res3, CA).pos - center;
  Matrix3d rotMatrix(axis, torsion1, center);
  
  p.atom(res1,C).pos.applyMatrix(rotMatrix);
  p.atom(res1,O).pos.applyMatrix(rotMatrix);

  for (int i=res1+1; i<res3; i++)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(rotMatrix);

  if (p.hasAtom(res3,H)) 
    p.atom(res3,H).pos.applyMatrix(rotMatrix);
  p.atom(res3,N).pos.applyMatrix(rotMatrix);

  Vector3d oldN2 = p.atom(res2,N).pos;
  Vector3d oldC2 = p.atom(res2,C).pos;

  // cary out second rotation between CA1 and CA2
  double torsion2 = pos_dihedral(p.atom(res1, C).pos,   p.atom(res1, CA).pos, 
                                 p.atom(res2, CA).pos,  newC1);
  center = p.atom(res1, CA).pos;
  axis = p.atom(res2, CA).pos - center;
  Matrix3d rotMatrix2(axis, torsion2, center);

  p.atom(res1,C).pos = newC1;
  p.atom(res1,O).pos.applyMatrix(rotMatrix2);

  for (int i=res1+1; i<res2; i++)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(rotMatrix2);

  if (p.hasAtom(res2,H)) 
    p.atom(res2,H).pos.applyMatrix(rotMatrix2);
  p.atom(res2,N).pos.applyMatrix(rotMatrix2);


  // calculate new phi1
  double dPhi1  = pos_dihedral(oldC1, p.atom(res1,N).pos, p.atom(res1,CA).pos, newC1);
  center = p.atom(res1,CA).pos;
  axis = center - p.atom(res1, N).pos;
  Matrix3d rotMatrix3(axis, dPhi1, center);

  p.atom(res1,HA1).pos.applyMatrix(rotMatrix3);
  for (AtomIter a = p.resAtomBegin(res1); a != p.resAtomEnd(res1); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(rotMatrix3);


  // cary out third rotation between CA2 and CA3
  double torsion3  = pos_dihedral(p.atom(res2, C).pos,   p.atom(res2, CA).pos, 
                                   p.atom(res3, CA).pos,  newC2);
  center = p.atom(res2, CA).pos;
  axis = p.atom(res3, CA).pos - center;
  Matrix3d rotMatrix5(axis, torsion3, center);

  p.atom(res2,C).pos = newC2;
  p.atom(res2,O).pos.applyMatrix(rotMatrix5);
  
  for (int i=res2+1; i<res3; i++)
    for (AtomIter a = p.resAtomBegin(i); a != p.resAtomEnd(i); a++)
      a->pos.applyMatrix(rotMatrix5);

  if (p.hasAtom(res3,H)) 
    p.atom(res3,H).pos.applyMatrix(rotMatrix5);
  p.atom(res3,N).pos.applyMatrix(rotMatrix5);


  // calculate orientation of res2 using oldC2 and oldN2
  center = p.atom(res2, CA).pos; 
  Vector3d old_ca_c = oldC2 - center;
  Vector3d new_ca_c = p.atom(res2,C).pos - center;

  axis = cross(old_ca_c, new_ca_c);
  double torsion4 = vec_dihedral(new_ca_c, axis, old_ca_c);
  
  Matrix3d rotMatrix6(axis, torsion4, center);  
  p.atom(res2,HA1).pos.applyMatrix(rotMatrix6);
  for (AtomIter a = p.resAtomBegin(res2); a != p.resAtomEnd(res2); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(rotMatrix6);
  oldN2.applyMatrix(rotMatrix6);

  Vector3d old_ca_n = oldN2 - center;
  Vector3d new_ca_n = p.atom(res2,N).pos - center;
    
  axis = new_ca_c;
  double torsion5 = vec_dihedral(new_ca_n, axis, old_ca_n);
  
  Matrix3d rotMatrix7(axis, torsion5, center);  
  p.atom(res2,HA1).pos.applyMatrix(rotMatrix7);
  for (AtomIter a = p.resAtomBegin(res2); a != p.resAtomEnd(res2); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(rotMatrix7);


  // calculate new psi3 angle
  double dPsi3  = pos_dihedral(oldN3, p.atom(res3,C).pos, 
                               p.atom(res3,CA).pos,  p.atom(res3,N).pos);
  center = p.atom(res3,CA).pos;
  axis = center - p.atom(res3, C).pos;
  Matrix3d rotMatrix8(axis, dPsi3, center);

  p.atom(res3,HA1).pos.applyMatrix(rotMatrix8);
  for (AtomIter a = p.resAtomBegin(res3); a != p.resAtomEnd(res3); a++)
    if (a->key >= CB)
      a->pos.applyMatrix(rotMatrix8);

  return;
}




