//
//  ****************************************************************************
//
//  display.h
//
//  (C) Thu Mar 29 2004 Bosco K. Ho 
//
//  ****************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//
//

#ifndef _display_h
#define _display_h

#include <math.h>
#include <string>
#include <vector>
#include <list>
#include <fstream>

#include <allegro.h>
#include "protein.h"
#include "pdbreader.h"
#include "vector3d.h"
#include "graphic3d.h"
#include "widget.h"
#include "bond.h"
#include "sturm.h"
#include "tripep_closure.h"

#define WHITE         255
#define LIGHT_GREY    248
#define GREY          212
#define DARK_GREY     204
#define BLACK         0
#define RED           1
#define BLUE          2
#define YELLOW        3
#define GREEN         4
#define DARK_GREEN    6
#define LIGHT_BLUE    7
#define ORANGE        8
#define LIGHT_ORANGE  9


// dimensions for ball and stick 
const double  DISPLAY_ATOM_RADIUS = 0.20;
const double  DISPLAY_LINE_WIDTH  = 0.06;

// the CA-CA distance limit for the sphere view
const double  CA_DIST             = 8.0;
const double  CA_DIST_SQ          = CA_DIST * CA_DIST;

// length of the scrollbar
const  int    SCROLLBAR_SIZE      = 30;

// action mode for left mouse clicks
const  int    PICK_RESIDUE        = 0;
const  int    PICK_DISTANCE       = 1;
const  int    PICK_ANCHOR         = 2;

// display mode for pProtein
const  int    SHOW_ALL            = 0;
const  int    SHOW_SEGMENT        = 1;
const  int    SHOW_SPHERE         = 2;

// display mode for pProtein
const  int    STEREO_NONE         = 0;
const  int    STEREO_CROSSEYE     = 1;
const  int    STEREO_WALLEYE = 2;


struct AtomMarker
{
  int      res;
  int      atom;
};


// This class mediates interaction from program with 3d view and
// proteins
class ProteinDisplay : public Widget
{
public:
  ProteinDisplay(Protein* p, int x, int y, int w, int h);

  bool isCentered;
  bool isHbond;
  bool isSidechain;
  bool isSolid;
  bool isPeptide;
  bool isHydrogen;
  
  int  showState;
  int  pickState;
  int  stereoState;

  int  viewLeftRes;
  int  viewRightRes;

  int  activeLeftRes;
  int  activeRightRes;

  int  focusRes;
  int  focusAtom;

  int  mNLoopSolution;
  int  mILoopSolution;

  int  anchorRes1;
  int  anchorRes2;
  int  anchorRes3;

  Protein*  pProtein;

  int   handle(int event);
  void  draw(BITMAP *b);
  void  resetFocus(void);
  void  resetFocusWithSameOrientation(void);
  void  initializeFocus(void);
  void  clearMeasure() { mMeasureMarker.clear(); };
  void  clearAnchor() { mNLoopSolution = 0; anchorRes1 = anchorRes2 = anchorRes3 = -1; };
  void  changeLoopSolution(int iSolution);
  void  setNewAnchor(void);
  void  changeAnchorPhiPsi(double phi, double psi);
  void  computeLoopClosureSolutions();
private:

  void computeCameraFromFocus(void);
  void countBonds();
  void countBondsBetweenRes(int i, int j);

  void projectLine(int i, int a, int j, int b, int color);
  void projectBond(int i, int a, int j, int b);
  void projectAtom(int res, int atom);
  void projectZoomSphere(int res, int atom, int color);
  void projectHighlightAtom(int res, int atom, int color, int highlight_size);
  void projectBlockedAtom(int res, int atom, int color, int highlight_size);
  void projectPeptidePlane(int i);
  void projectMeasure();
  void projectAnchor();
  void projectResidue(int i);
  void projectBonds(int i);
  void projectAtoms(int i);
  bool resInSphere(int i, Vector3d& focus);
  
  void projectProteinToGraphicsView();

  // orientation for camera relative to pProtein
  Vector3d       mFocusPos;
  Vector3d       mFocusUp;
  Vector3d       mFocusAxis;
                 
  Vector3d       mCameraLook;
  Vector3d       mCameraUp;
                 
  int            mHighlightRes;
  int            mHighlightAtom;
  int            mHighlightZ;
  std::string    mHighlightString;

  // atoms to be measured
  std::vector<AtomMarker> 
                 mMeasureMarker;
  std::string    mMeasureDihedString;
  std::string    mMeasureDistString;
  std::string    mMeasureAngleString;
  std::string    mMeasureAngle2String;
  std::string    mMeasureDist2String;
  
  double r_soln_n[MAX_LOOP_SOLN][3][3];
  double r_soln_a[MAX_LOOP_SOLN][3][3];
  double r_soln_c[MAX_LOOP_SOLN][3][3];

  GraphicsView   mGraphicsView;
                 
  int            mSaveMouseY;
  int            mSaveMouseX;
                 
  int            mSavePickRes;
  int            mSavePickAtom;
                 
  HBondList      mHBondList;
  WeakHBondList  mWeakHBondList;
  DisulfideList  mDisulfideList;
};


#endif
