//
//  ***************************************************************************
//
//  graphics.cpp
//  (C) 2004 Bosco K. Ho 
//
//  GraphicsView handles Vector3d inputs, sorts by depth in 3d, and converts 
//  it into a 2d allegro bitmap. Requires camera information.
//
//  ***************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//

#include "graphic3d.h"

//******** GfxList ********************


void GfxList::calculateMatrix(Vector3d& pos, Vector3d& front, Vector3d& up)
{
  get_camera_matrix(
    &mProjectionMatrix, 
    ftofix(pos.x()),   ftofix(pos.y()),   ftofix(pos.z()), 
    ftofix(front.x()), ftofix(front.y()), ftofix(front.z()), 
    ftofix(up.x()),    ftofix(up.y()),    ftofix(up.z()), 
    itofix(38),        itofix(1));
  mCameraPos   = pos;
  mCameraFront = front;
  mCameraUp    = up.perpendicular(front);
}



void GfxList::getTransform(Vector3d& v, double r, 
                    int *xOut, int *yOut, int *zOut, int *radius)
{ 
  fixed x, y, z;
  projectPoint (v, &x, &y, &z);
  if (z <= itofix (1)) return;

  *xOut = fixtoi(x);
  *zOut = fixtoi(z);
  *yOut = fixtoi(y);

   fixed x_r, y_r, z_r;
  projectPoint ( v + mCameraUp.scaled(r), &x_r, &y_r, &z_r);
  if (z_r <= itofix (1)) return;
  *radius = fixtoi(fixhypot(x - x_r, y - y_r));
}



void GfxList::projectPoint (Vector3d p, fixed *xOut, fixed *yOut, fixed *zOut)
{
  fixed x, y, z;

  apply_matrix(&mProjectionMatrix, 
    ftofix (p.x ()), ftofix (p.y ()), ftofix (p.z ()),   &x, &y, &z); 
  persp_project(x, y, z, xOut, yOut);
  *zOut = z;
  *xOut += ftofix (mXOffset);
}



void GfxList::push_back(GfxObject& g) 
{
  mGfxObjectList.push_back(g);
}



void GfxList::clear() 
{
  mGfxObjectList.clear();
}


inline bool comp(const GfxObject& a, const GfxObject& b) { return a.z > b.z; };

void GfxList::sort () 
{ 
  mGfxObjectList.sort();
}



void GfxList::draw(BITMAP *bmp) 
{
  for (GfxObjectIter i = mGfxObjectList.begin(); i != mGfxObjectList.end(); i++) 
  {
    switch (i->id) 
    {
      case GFX_CIRCLEFILL:
        circlefill (bmp, i->v[0], i->v[1], i->v[2], i->color);
        break;
      case GFX_LINE:
        line (bmp, i->v[0], i->v[1], i->v[2], i->v[3], i->color);
        break;
      case GFX_CIRCLE:
        circle (bmp, i->v[0], i->v[1], i->v[2], i->color);
        break;
      case GFX_CIRCLE_OUTLINE :
        circlefill (bmp, i->v[0], i->v[1], i->v[2], i->color);
        circle (bmp, i->v[0], i->v[1], i->v[2]+1, OUTLINE_COLOR);
        break;
      case GFX_QUADFILL:
        polygon (bmp, 4, i->v, i->color);
        break;
      case GFX_QUAD:
        wireframe (bmp, 4, i->v, i->color);
        break;
      case GFX_QUAD_OUTLINE:
        polygon (bmp, 4, i->v, i->color);
        wireframe (bmp, 4, i->v, OUTLINE_COLOR);
        break;
      case GFX_QUAD_WIREFRAME:
        wireframe (bmp, 4, i->v, i->color);
        break;
      case GFX_TEXTSTRING:
        stringout (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1] - 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0]    , i->v[1] - 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1]    , OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1] + 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0]    , i->v[1] + 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1]    , OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1] + 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1] - 1, OUTLINE_COLOR);
        stringout (bmp, font, *(i->pStr), i->v[0]    , i->v[1]    , i->color    );
        break;
      case GFX_TEXTSTRING_CENTRE:
        stringout_centre (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1] - 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0]    , i->v[1] - 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1]    , OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1] + 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0]    , i->v[1] + 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1]    , OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0] - 1, i->v[1] + 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0] + 1, i->v[1] - 1, OUTLINE_COLOR);
        stringout_centre (bmp, font, *(i->pStr), i->v[0]    , i->v[1]    , i->color    );
        break;
    }
  }
}



//  Projecting 3d primitives into 2d primitives



void GfxList::projectThickLine(Vector3d& p1, Vector3d& p2, int color, double width) 
{
  fixed x1, y1, z1;
  projectPoint(p1, &x1, &y1, &z1);
  if (z1 <= itofix (1)) return;

  fixed x1_r, y1_r, z1_r;
  projectPoint(p1 + mCameraUp.scaled(width), &x1_r, &y1_r, &z1_r);
  if (z1_r <= itofix (1)) return;

  fixed r1 = fixhypot(x1 - x1_r, y1 - y1_r);

  fixed x2, y2, z2;
  projectPoint(p2, &x2, &y2, &z2);
  if (z2 <= itofix (1)) return;

  fixed x2_r, y2_r, z2_r;
  projectPoint(p2 + mCameraUp.scaled(width), &x2_r, &y2_r, &z2_r);
  if (z2_r <= itofix (1)) return;

  fixed r2 = fixhypot(x2 - x2_r, y2 - y2_r);
  
  fixed dx = - (y2 - y1);
  fixed dy = x2 - x1;
  fixed d  = fixhypot(dx, dy);

  dx = fixdiv(dx, d);
  dy = fixdiv(dy, d);

  GfxObject gfxObject;
  gfxObject.id    = GFX_QUADFILL;
  gfxObject.v[0]  = fixtoi(x1 + fixmul(dx, r1));
  gfxObject.v[1]  = fixtoi(y1 + fixmul(dy, r1));
  gfxObject.v[2]  = fixtoi(x2 + fixmul(dx, r2));
  gfxObject.v[3]  = fixtoi(y2 + fixmul(dy, r2));
  gfxObject.v[4]  = fixtoi(x2 - fixmul(dx, r2));
  gfxObject.v[5]  = fixtoi(y2 - fixmul(dy, r2));
  gfxObject.v[6]  = fixtoi(x1 - fixmul(dx, r1));
  gfxObject.v[7]  = fixtoi(y1 - fixmul(dy, r1));
  gfxObject.color = color;
  if (z1 > z2) 
    gfxObject.z   = fixtoi(z1); 
  else
    gfxObject.z   = fixtoi(z2); 
  push_back(gfxObject);

  r1 += itofix (1);
  r2 += itofix (1);

  gfxObject.id    = GFX_QUAD;
  gfxObject.v[0]  = fixtoi(x1 + fixmul(dx, r1));
  gfxObject.v[1]  = fixtoi(y1 + fixmul(dy, r1));
  gfxObject.v[2]  = fixtoi(x2 + fixmul(dx, r2));
  gfxObject.v[3]  = fixtoi(y2 + fixmul(dy, r2));
  gfxObject.v[4]  = fixtoi(x2 - fixmul(dx, r2));
  gfxObject.v[5]  = fixtoi(y2 - fixmul(dy, r2));
  gfxObject.v[6]  = fixtoi(x1 - fixmul(dx, r1));
  gfxObject.v[7]  = fixtoi(y1 - fixmul(dy, r1));
  gfxObject.color = OUTLINE_COLOR;
  push_back(gfxObject);
}


void GfxList::projectThinLine(Vector3d& p1, Vector3d& p2, int color) 
{
  fixed x1, y1, z1;
  projectPoint(p1, &x1, &y1, &z1);
  if (z1 <= itofix (1)) return;

  fixed x2, y2, z2;
  projectPoint(p2, &x2, &y2, &z2);
  if (z2 <= itofix (1)) return;

  GfxObject gfxObject;
  gfxObject.id    = GFX_LINE;
  gfxObject.v[0]  = fixtoi(x1);
  gfxObject.v[1]  = fixtoi(y1);
  gfxObject.v[2]  = fixtoi(x2);
  gfxObject.v[3]  = fixtoi(y2);
  gfxObject.color = color;
  if (z1>z2) 
    gfxObject.z   = fixtoi (z1); 
  else
    gfxObject.z   = fixtoi (z2); 
  push_back(gfxObject);
}



void GfxList::projectString(Vector3d& p1, int color, std::string* pStr) 
{
  fixed x1, y1, z1;
  projectPoint (p1, &x1, &y1, &z1);
  if (z1 <= itofix (1)) return;

  GfxObject gfxObject;
  gfxObject.id    = GFX_TEXTSTRING;
  gfxObject.v[0]  = fixtoi(x1) + 7;
  gfxObject.v[1]  = fixtoi(y1) - 3;
  gfxObject.color = color;
  gfxObject.z     = fixtoi(0); 
  gfxObject.pStr  = pStr;
  push_back(gfxObject);
}



void GfxList::projectCentreString(Vector3d& p1, int color, std::string* pStr) 
{
  fixed x1, y1, z1;
  projectPoint (p1, &x1, &y1, &z1);
  if (z1 <= itofix (1)) return;

  GfxObject gfxObject;
  gfxObject.id    = GFX_TEXTSTRING_CENTRE;
  gfxObject.v[0]  = fixtoi(x1) + 2;
  gfxObject.v[1]  = fixtoi(y1) - 4;
  gfxObject.color = color;
  gfxObject.z     = fixtoi(0); 
  gfxObject.pStr  = pStr;
  push_back(gfxObject);
}



void GfxList::projectQuad (Vector3d v[], int color)
{
  GfxObject gfxObject;

  gfxObject.color = color;
  gfxObject.id    = GFX_QUAD_OUTLINE;
  gfxObject.z     = 0;
  for (int i=0; i<=3; i+=1) 
  {
    fixed x, y, z;
    projectPoint (v[i], &x, &y, &z);
    if (z <= itofix (1)) return;
    gfxObject.v[i*2]     = fixtoi (x);
    gfxObject.v[i*2 + 1] = fixtoi (y);
    if (fixtoi (z) > gfxObject.z) 
      gfxObject.z = fixtoi (z);
  }
  push_back(gfxObject);
}



void GfxList::projectWireframeQuad (Vector3d v[], int color)
{
  GfxObject gfxObject;

  gfxObject.color = color;
  gfxObject.id    = GFX_QUAD_WIREFRAME;
  gfxObject.z     = 0;
  for (int i=0; i<=3; i+=1) 
  {
    fixed x, y, z;
    projectPoint (v[i], &x, &y, &z);
    if (z <= itofix (1)) return;
    gfxObject.v[i*2]     = fixtoi (x);
    gfxObject.v[i*2 + 1] = fixtoi (y);
    if (fixtoi (z) > gfxObject.z) 
      gfxObject.z = fixtoi (z);
  }
  push_back(gfxObject);
}


void GfxList::projectFilledCircle (Vector3d& p, double r, int color)
{
  fixed x, y, z;
  projectPoint (p, &x, &y, &z);
  if (z <= itofix (1)) return;

   fixed x_r, y_r, z_r;
  projectPoint ( p + mCameraUp.scaled(r), &x_r, &y_r, &z_r);
  if (z_r <= itofix (1)) return;

  fixed radius = fixhypot (x - x_r, y - y_r);

  GfxObject gfxObject;

  gfxObject.id    = GFX_CIRCLE_OUTLINE;
  gfxObject.v[0]  = fixtoi(x);
  gfxObject.v[1]  = fixtoi(y);
  gfxObject.v[2]  = fixtoi(radius);
  gfxObject.color = color;
  gfxObject.z     = fixtoi(z);
  push_back(gfxObject);
}



void GfxList::projectEnlargedCircle (Vector3d& p, double r, int enlargedRad, int color)
{
  fixed x, y, z;
  projectPoint (p, &x, &y, &z);
  if (z <= itofix (1)) return;

  fixed x_r, y_r, z_r;
  projectPoint (p + mCameraUp.scaled(r), &x_r, &y_r, &z_r);
  if (z_r <= itofix (1)) return;

  fixed radius = fixhypot (x-x_r, y-y_r);

  GfxObject gfxObject;

  gfxObject.id=GFX_CIRCLE;
  gfxObject.v[0]=fixtoi (x);
  gfxObject.v[1]=fixtoi (y);
  gfxObject.v[2]=fixtoi (radius) + enlargedRad + 1;
  gfxObject.color=color;
  gfxObject.z=fixtoi (z);
  push_back(gfxObject);

  gfxObject.id=GFX_CIRCLE;
  gfxObject.v[2]=fixtoi (radius) + enlargedRad;
  push_back(gfxObject);

}



void GfxList::projectEnlargedFilledCircle (Vector3d& p, double r, int enlargedRad, int color)
{
  fixed x, y, z;
  projectPoint (p, &x, &y, &z);
  if (z <= itofix (1)) return;

  fixed x_r, y_r, z_r;
  projectPoint (p + mCameraUp.scaled(r), &x_r, &y_r, &z_r);
  if (z_r <= itofix (1)) return;

  fixed radius = fixhypot (x-x_r, y-y_r);

  GfxObject gfxObject;

  gfxObject.id=GFX_CIRCLE_OUTLINE;
  gfxObject.v[0]=fixtoi (x);
  gfxObject.v[1]=fixtoi (y);
  gfxObject.v[2]=fixtoi (radius) + enlargedRad + 1;
  gfxObject.color=color;
  gfxObject.z=fixtoi (z);
  push_back(gfxObject);
}


//  GraphicsView



void GraphicsView::projectThickLine (Vector3d& p1, Vector3d& p2, int color, double width)
{
  gfxList.projectThickLine(p1, p2, color, width);
  if (isStereo)
    gfxListR.projectThickLine(p1, p2, color, width);
}



void GraphicsView::projectThinLine (Vector3d& p1, Vector3d& p2, int color)
{
  gfxList.projectThinLine(p1, p2, color);
  if (isStereo)
    gfxListR.projectThinLine(p1, p2, color);
}



void GraphicsView::projectString (Vector3d& p1, int color, std::string* pStr)
{
  gfxList.projectString(p1, color, pStr);
  if (isStereo)
    gfxListR.projectString(p1, color, pStr);
}



void GraphicsView::projectCentreString (Vector3d& p1, int color, std::string* pStr)
{
  gfxList.projectCentreString(p1, color, pStr);
  if (isStereo)
    gfxListR.projectCentreString(p1, color, pStr);
}



void GraphicsView::projectQuad (Vector3d v[], int color)
{
  gfxList.projectQuad(v, color);
  if (isStereo)
    gfxListR.projectQuad(v, color);
}


void GraphicsView::projectWireframeQuad (Vector3d v[], int color)
{
  gfxList.projectWireframeQuad(v, color);
  if (isStereo)
    gfxListR.projectWireframeQuad(v, color);
}



void GraphicsView::projectFilledCircle (Vector3d& p, double r, int color)
{
  gfxList.projectFilledCircle(p, r, color);
  if (isStereo)
     gfxListR.projectFilledCircle(p, r, color);
}



void GraphicsView::projectEnlargedCircle (Vector3d& p, double r, int enlargedRad, int color)
{
  gfxList.projectEnlargedCircle(p, r, enlargedRad, color);
  if (isStereo)
    gfxListR.projectEnlargedCircle(p, r, enlargedRad, color);
}



void GraphicsView::projectEnlargedFilledCircle (Vector3d& p, double r, int enlargedRad, int color)
{
  gfxList.projectEnlargedFilledCircle(p, r, enlargedRad, color);
  if (isStereo)
    gfxListR.projectEnlargedFilledCircle(p, r, enlargedRad, color);
}



bool GraphicsView::isInCircle(Vector3d& v, double racceptorut, int testX, int testY, int *zOut)
{
  int x, y, r;
  gfxList.getTransform(v, racceptorut, &x, &y, zOut, &r);

  if ((testX >= x-r-1) && (testX <= x+r+1)
    && (testY >= y-r-1) && (testY <= y+r+1))
  {
    return true;
  }

  if (! isStereo) return false;

  gfxListR.getTransform(v, racceptorut, &x, &y, zOut, &r);
  if ((testX >= x-r-1) && (testX <= x+r+1)
    && (testY >= y-r-1) && (testY <= y+r+1))
  {
    return true;
  }

  return false;
}


void GraphicsView::draw(BITMAP *bmp)
{
  /* draw left eye p.o.v. if stereo */
  if (isStereo)
    set_clip(bmp, bmp->w/2 - stereoSeparation - 10, 0, bmp->w/2-10, bmp->h);

  gfxList.sort();
  gfxList.draw(bmp);

  if (isStereo)
  {
    /* draw right eye p.o.v. */
    set_clip(bmp, bmp->w/2+10, 0, bmp->w/2 + stereoSeparation+10, bmp->h);
    gfxListR.sort();
    gfxListR.draw(bmp);

    set_clip(bmp, 0, 0, bmp->w, bmp->h);
  }
}


void GraphicsView::calculateMatrix(Vector3d& pos, Vector3d& front, Vector3d& up)
{
  set_projection_viewport(0, 0, SCREEN_W, SCREEN_H);

  gfxList.calculateMatrix(pos, front, up);
  gfxList.setXOffset(0);

  if (!isStereo) return;

  gfxList.setXOffset(-stereoSeparation/2 - 11);

  Vector3d front_r = front;

  front_r.rotate(-stereoViewAngle, up);

  Vector3d pos_r = pos + front - front_r;

  gfxListR.calculateMatrix(pos_r, front_r, up);
  gfxListR.setXOffset(stereoSeparation/2 + 10);

}

