//
//  ****************************************************************************
//
//  graphic3d.h
//
//  (C) Thu Mar 25 2004 Bosco K. Ho 
//
//  ****************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//
//

#ifndef _graphic3d_h_
#define _graphic3d_h_

#include <allegro.h>
#include <list>
#include "vector3d.h"
#include "util.h"


// graphical objects for display //
typedef enum 
{
  GFX_QUADFILL,
  GFX_CIRCLEFILL,
  GFX_CIRCLE,
  GFX_QUAD,
  GFX_LINE,
  GFX_TEXTSTRING,
  GFX_TEXTSTRING_CENTRE,
  GFX_QUAD_OUTLINE,
  GFX_QUAD_WIREFRAME,
  GFX_CIRCLE_OUTLINE
};


struct GfxObject 
{
  int          id;
  int          v[8];
  int          z;
  std::string* pStr;
  int          color;
};
inline bool operator < (const GfxObject& a, const GfxObject& b)  { return (a.z > b.z);  }
// must use list for the fastest sorting
typedef std::list<GfxObject>    GfxObjectList;
typedef GfxObjectList::iterator GfxObjectIter;
 

class GfxList
{
  public:
    GfxList() {}
    ~GfxList() {}

    void clear();
    void sort();
    void draw(BITMAP* bmp);

    void projectThickLine(Vector3d& p1, Vector3d& p2, int color, double width);
    void projectThinLine(Vector3d& p1, Vector3d& p2, int color);
    void projectString (Vector3d& p1, int color, std::string* pStr);
    void projectCentreString (Vector3d& p1, int color, std::string* pStr);
    void projectQuad(Vector3d v[], int color);
    void projectWireframeQuad(Vector3d v[], int color);
    void projectFilledCircle(Vector3d& p, double r, int color);
    void projectEnlargedCircle(Vector3d& p, double r, int enlargedRad, int color);
    void projectEnlargedFilledCircle(Vector3d& p, double r, int enlargedRad, int color);

    void setXOffset(int x) { mXOffset = x; }
    void calculateMatrix(Vector3d& pos, Vector3d& front, Vector3d& up);
    void getTransform(Vector3d& v, double r,
                int *xOut, int *yOut, int *zOut, int *radius);

  private:
    void projectPoint(Vector3d p, fixed *xOut, fixed *yOut, fixed *zOut);
    void push_back(GfxObject& g);
 
    GfxObjectList mGfxObjectList;
    MATRIX        mProjectionMatrix;
    int           mXOffset;
    Vector3d      mCameraPos;
    Vector3d      mCameraFront;
    Vector3d      mCameraUp;
};





// this class produces a stereo display 
class GraphicsView
{
  public:
    GraphicsView() : isStereo(false), stereoViewAngle(-8.0), stereoSeparation(300) {};
    ~GraphicsView() {};

   // parameters for stereo
    bool    isStereo;
    double  stereoViewAngle;
    int     stereoSeparation;

    GfxList gfxList;
    GfxList gfxListR;

    void clear() { gfxList.clear(); if (isStereo) gfxListR.clear(); };
    void sort()  { gfxList.sort();  if (isStereo) gfxListR.sort(); };
    void draw(BITMAP* bmp);
    void calculateMatrix(Vector3d& pos, Vector3d& front, Vector3d& up);

    void projectThickLine (Vector3d& p1, Vector3d& p2, int color, double width);
    void projectThinLine (Vector3d& p1, Vector3d& p2, int color);
    void projectString (Vector3d& p1, int color, std::string* pStr);
    void projectCentreString (Vector3d& p1, int color, std::string* pStr);
    void projectQuad (Vector3d v[], int color);
    void projectWireframeQuad(Vector3d v[], int color);
    void projectFilledCircle (Vector3d& p, double r, int color);
    void projectEnlargedCircle (Vector3d& p, double r, int enlargedRad, int color);
    void projectEnlargedFilledCircle(Vector3d& p, double r, int enlargedRad, int color);

    bool isInCircle(Vector3d& v, double r, int testX, int testY, int *zOut);

    void setStereoSeparation(int separation) { stereoSeparation = separation; };
    void setStereoViewAngle(float viewAngle) { stereoViewAngle = viewAngle; };
};


#endif
