//
//  **************************************************************************
//
//  polymer.cpp
//
//  (C) 2003 Bosco K. Ho 
//
//  The purpose of the Polymer class is to provide a convenient 
//  structure to access the atoms via the Resblk structures without 
//  disturbing the underlying structure of the atoms.
//  The Res structure allows you to access the atom by 
//  residue number and atom-type. 
//
//  There are basically three categories of routines
//  
//  a) routines to read a pdb file and extract dihedral angles
//  b) routines to rotate and change internal parameters of a loaded
//     pdb file
//  c) routines to generate the residues de novo
//  d) various diagnostics to check protein integrity
//
//  In terms of de-novo strucutre determination, the information
//  to generate the structures are kept in the code instead of using
//  extensive data structures, I find this is easier to read and 
//  highlights the similarity between various amino acids.
//
//  Another important question is how do we index the residues and
//  chi angles? In C++, one assumes the count to start from 0 whereas
//  in most other cases, one starts counting from one.
//
//  Adapted from the The Biomolecule Toolkit (C) 2001, Eric Alm 
//
//  Cleaned the abstraction interface. Introduced a key for atoms
//  as well as for amino acids. This avoids the clumsy method of
//  using strings to identify atoms. There is too much confusion in the
//  peptide class since there is much interface between the protein
//  and peptide class. Consequently, the peptide class has been changed
//  to a lookup table class with little intrinsic functionality. 
//  Got rid of templates.
//
//
//  24/2/2004 fixed the HA1 and HA2 hydrogen sprout
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#include <deque>
#include <list>
#include "polymer.h"
#include "vector3d.h"

using namespace std;



AtomIter& Polymer::resAtomBegin(int i)
{ 
  if (nAtomInRes(i) == 0)
    return mResList[i].atomEnd;
  else
    return *(mResList[i].atomIterList.begin()); 
};



AtomIter& Polymer::resAtomEnd(int i)
{ 
  return mResList[i].atomEnd;
};



// copy constructor
Polymer::Polymer(const Polymer& rhs) 
{
  clear();
  insertFragment(rhs, 0);
}



// -------- access functions



bool Polymer::hasAtom(int i, int atomKey) 
{
  if ( (i < 0) || (i > nRes() - 1) ) return false;

  for (int j = 0; j < nAtomInRes(i); j++)
    if (mResList[i].atomIterList[j]->key == atomKey) 
      return true;

  return false;
}



int Polymer::nRes() const
{ 
  return mResList.size(); 
}



Atom& Polymer::atom(int i, int atomKey)
{
  for (int j = 0; j < nAtomInRes(i); j++)
  {
    if (mResList[i].atomIterList[j]->key == atomKey)
      return *(mResList[i].atomIterList[j]);
  }
  return NULL_ATOM;
}


// constructs a string of the form 'LYS:32-CA'
string Polymer::resAtomName(int i, int a)
{
  ostringstream os;
  os << resName(i) << ":" << resNo(i); 
  string aName = atom(i, a).name;
  while (' ' == aName[0]) aName.erase(0,1);
  return os.str() + "-" + aName; 
}



double Polymer::maxLength(void)
{
  double lengthSquared = 0.0;
  double newLengthSquared;
  for (int i = 0; i < nRes(); i++)
    for (int ia = 0; ia < nAtomInRes(i); ia++)
      for (int j = i+1; j < nRes(); j++)
        for (int ja = 0; ja < nAtomInRes(j); ja++)
        {
          Vector3d& p1 = mResList[i].atomIterList[ia]->pos;
          Vector3d& p2 = mResList[j].atomIterList[ja]->pos;
          newLengthSquared = pos_dist_squared(p1, p2);
          if (newLengthSquared > lengthSquared)
            lengthSquared = newLengthSquared;
        }
  return sqrt(lengthSquared);
}



// -------- insertion/deletion of atoms 



// inserts atom into the molecule and updates residue block
void Polymer::insertAtom(int i, const Atom& atom) 
{
  Atom a = atom;
  a.resKey              = resKey(i); 
  a.resName             = resName(i); 
  a.resNo               = resNo(i); 
  AtomIter& insertPoint = resAtomEnd(i);

  AtomIter newAtomIter = mAtomList.insert(insertPoint, a);
  mResList[i].atomIterList.push_back(newAtomIter);

  // if resBlk is filled for the first time, must set atomEnd
  // of prev resBlk
  if (nAtomInRes(i) == 1)
    setAtomEnd(i-1, resAtomBegin(i));
}


  
// inserts atom into the molecule using a key
void Polymer::insertAtom(int i, int atomKey) 
{
  Atom a;
  a.key  = atomKey; 
  a.name = getAtomName(resKey(i), atomKey); 
  insertAtom(i, a);
}



// inserts atom into the molecule and updates residue block
void Polymer::eraseAtom(int i, int atomKey) 
{
  typedef AtomIterList::iterator AtomIterIter;
  AtomIterIter iter;
  AtomIterIter iterBegin = mResList[i].atomIterList.begin();
  AtomIterIter iterEnd   = mResList[i].atomIterList.end();
  for (iter = iterBegin; iter != iterEnd; iter++) 
  {
     AtomIter a = *iter;
     if (a->key == atomKey) 
     {
       mResList[i].atomIterList.erase(iter);
       mAtomList.erase(a);
       // if the first atom and/or very last atom in
       // the Res is being deleted
       if (iter == iterBegin)
         setAtomEnd(i-1, resAtomBegin(i));
       return;
     }
  }
}



void Polymer::setAtomEnd(int i, AtomIter a) 
{
  if (i < 0) 
    return;
  mResList[i].atomEnd = a;
  if ( (i > 0) && (nAtomInRes(i) == 0) )
    setAtomEnd(i-1, a);
}



void Polymer::sproutAtom(int i, int a1, double length, double angle, double dihed,
                                 int a2, int a3, int a4) 
{
  if (hasAtom(i,a1))
    atom(i, a1).pos  = project_pos(length, angle, dihed,
              atom(i, a2).pos, atom(i, a3).pos, atom(i, a4).pos);
}


// constructs the resBlk entries from the flat atom structures
// in the parent Molecule class
void Polymer::scanRes() 
{
  mResList.clear();

  int thisResNo = atomBegin()->resNo; 
  int i = -1;
  Res resBlk;
  for (AtomIter a = atomBegin(); a != atomEnd(); a++) 
  {
    // check when residue number changes
    if ( (a->resNo != thisResNo) || (i == -1) ) 
    { 
      // first set the atomEnd of the previous residue
      if (i >= 0) 
        mResList[i].atomEnd = a;

      thisResNo      = a->resNo;
      resBlk.no   = a->resNo;  
      resBlk.key  = a->resKey;  
      resBlk.name = a->resName;  

      mResList.push_back(resBlk);
      ++i;
    }
    mResList[i].atomIterList.push_back(a);
  }
  if (i >= 0) 
    mResList[i].atomEnd = atomEnd();
}



void Polymer::insertRes(int i, const Res& r)
{
  if ( (i < 0) || (i > nRes()) ) return;
  
  mResList.push_back(r); 

  // copies every residue from i to i+1 and updates resNo 
  for (int j = nRes() - 1; j > i; j--) 
  {
    mResList[j] = mResList[j-1];
    mResList[j].no += 1;
    for (int k = 0; k < nAtomInRes(j); k++)
      mResList[j].atomIterList[k]->resNo += 1;
  }

  for (int k = 0; k < MAX_RES_PARAM; k++)
    mResList[i].param[k]  = r.param[k];

  // sets the appropriate atomEnd
  if (i == nRes() - 1)
  {
    mResList[i].atomEnd = atomEnd();
  }
  else
  {
    // makes sure that atomEnd matches the next resBlk 
    mResList[i].atomEnd = resAtomBegin(i+1);
  }
  
  if (nRes() == 1)
    mResList[i].no = 1;
  else
  {
    if (i == 0)
      mResList[i].no = mResList[i+1].no - 1;
    else
      mResList[i].no = mResList[i-1].no + 1;
  }
  mResList[i].key = r.key;
  mResList[i].name = r.name;
  mResList[i].atomIterList.clear();

  for (unsigned int k = 0; k < r.atomIterList.size(); k++)
  {
    Atom a = *(r.atomIterList[k]);
    insertAtom(i, a);
  }

  // makes sure that the prev atomEnd is
  // pointing in the right place
  if (i > 0)
    setAtomEnd(i-1, resAtomBegin(i));
}



void Polymer::eraseRes(int i) 
{
  if (i > nRes() - 1) return;

  AtomIter firstAtomIter = resAtomBegin(i);
  AtomIter lastAtomIter  = resAtomEnd(i);
  mAtomList.erase(firstAtomIter, lastAtomIter);
  
  for (int j = i; j < nRes() - 1; j++) 
  {
    mResList[j] = mResList[j+1];      
    mResList[j].no -= 1;      
    for (int k = 0; k < nAtomInRes(j); k++)
      mResList[j].atomIterList[k]->resNo -= 1;
  }
  mResList.pop_back();

  if (i > 0) 
    setAtomEnd(i-1, lastAtomIter);
} 



void Polymer::extractFragment(Polymer& p, int i, int j)
{
  p.clear();
  for (int k = j; k >= i; k--)
    p.insertRes(0, mResList[k]);
}



void Polymer::insertFragment(const Polymer& p, int i)
{
  for (int j = p.nRes() - 1; j >= 0; j--)
    insertRes(i, p.mResList[j]);
}

