//
//  **************************************************************************
//
//  The Ramachandran-plot Explorer(C) 2003 Bosco K. Ho
//
//  rama.cpp
//
//  This is a gProtein viewer. 
//  The emphasis of this program is on interactivity with the gProtein.
//  Easily see and *change* phi/psi/chi angles
//  Add, cut and paste gProtein fragments: allows for
//  fast profiling of gProtein conformations
//
//  Uses the allegro games library for graphics and input
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or(at
//  your option) any later version.
//
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
/*
  
    To do list
    ==========
    
        - extend to multiple chains and hetereo atoms
        - fix the standard initialize atom list
        - bond lists
        - clipping bug
        - states of histidine/get rid of titration
  
    Development log
    ===============
    
    25/12/2003:
        - added comments and license text
    27/12/2003:
        - fixed the copy residue bug
        - added pop-mCameraUp text when mouse is over atom
        - stream-lined vector library, including the rotation and center
          transform
        - accurate resNo accounting in insert/delete operations
        - simplified the pdbReader code
    29/12/2003:
        - modularised the 3d graphics class into a separate file
        - added a full atom name method
    30/12/2003:
        - create WidgetList class
    7/3/2004:
        - found out-of-bounds array read in the drawing of
          the peptide C-N bond. the for list counted one too many
          and the gProtein::hasAtom procedure did not check
          residue index bounds
    11/3/2004:
        - fixed load and save dialogs so that they don't crash
          with bad input
        - removed the ESC exit, added a quit button
        - on loading a new gProtein, will find the closest CA atom
          to the center of mass and do a zoom to showState the whole gProtein
        - will evaluate sidechain-mainchain h-bonds within a residue
        - prevent H atom projects on non-amino acid residues
        - changes the window title to new save file name in window
          title
     4/4/2004:
        - self h-bond evaluator
        - remembers to update resKey and resName on mutation
        - avoids projection of H and drawing of peptide bond 
          for non-standard amino acids
        - now changes name to generic in title when file is saved
        - fix bug: exits if loaded file is empty
        - added argv to open the file if argument is supplied
        - major refactor to separate gProtein display from widget from
          general gProtein manupilation routines
        - added a highlight text to describe atom when mouse is over
          atom
        - added a routine to quickly center and zoom gProtein on load
          and also adjusts scroll-bar correctly
        - added half-bar for green and pick in scrollbar
        - when dragging active residues in sequence box, can go off limits
          and scrolls
        - added the H2, H3, OXT, and HXT atoms
        - now allows 3 stereo modes
        - now does cannonical h-bonds (orange), weak h-bonds (pink)
          and disulfide bonds (yellow)
     6/14/2004:
        - fixed atom name differences between amino acids
        - fixed bug with delete for blocking groups
     6/30/2004:
        - cut and paste
        - blocking group options
     7/5/2004:
        - file in and out for cut-and-paste
        - cleaned up some names
        - phi/psi angles for n and c terminal residues
     7/11/2004:
        - make all arrays go from 0 to n-1, including the sidechain
        - eliminated the useless protDisp object, originally designed
          for multiple chains, but now abandoned
     7/14/2004:
        - nice big sphere for sphere-view
     8/5/2004:
        - introduced a polymer class with no references to proteins
        - fixed the initializing focus bug
        - found subtle view-port resize bug
        - made highlighted atoms all green
        - found a bug with the insertRes routine
        - fixed the cut-and-paste for consistency
        - made selecting atoms by clicking on the button release
          and double checks if the atom chosen is the same
          as button press
        - make name labels float next to atom
        - no more c-style strings!!! a cause to celebrate
        - renamed variables using the pseudo-hungarian style m, r, and p
        - measure tool
        - universal file dumping
        - added an end marker
        - added tick marks and secondary structure labels
        - charged ends/blocked
        - for show-sphere, checks *every* atom for inclusion in sphere
      01/17/2005
        - added proline coupling to backbone
        - added anchor for loop exploration
      
  
  **************************************************************************/


#include <string>
#include <allegro.h>
#include <iomanip>
#include <fstream>
#include "display.h"
#include "residue.h"
#include "control.h"


using namespace std;


bool            gIsProgramEnd = false;
bool            gIsAnimate = false;
Protein         gProtein;
ProteinDisplay* gpProteinDisplay;
WidgetList      gWidgetList;


const string VERSION_STRING =
  "The Ramachandran Plot Explorer V0.9 (c) 2005 Bosco Ho";



void handle_close()
{
  gIsProgramEnd = true;
}



void animate_protein()
{
  const double PHI1 = -150.0;
  const double PSI1 = 120.0;
  const double PHI2 = -60.0;
  const double PSI2 = -40.0;
  const double    d = sqrt( (PHI1-PHI2)*(PHI1-PHI2) + (PSI1-PSI2)*(PSI1-PSI2) );
  const double dphi = (PHI2-PHI1) / d;
  const double dpsi = (PSI2-PSI1) / d;

  for (int c = 0; c < gProtein.nRes(); c++)
  {
    double phi = gProtein.phi(c) + dphi;
    double psi = gProtein.psi(c) + dpsi;
    normalize_angle(phi);
    normalize_angle(psi);
    gProtein.phi(c) = phi;
    gProtein.psi(c) = psi;
    for (int d = 0; d < getNChi(gProtein.resKey(c)); d++)
    {
      double chi = gProtein.chi(c,d) + dphi * 2;
      normalize_angle(chi);
      gProtein.chi(c, d) = chi;
    }
  }
}



void reset_point_of_view()
{
  gpProteinDisplay->resetFocusWithSameOrientation();
}



void build_default_protein()
{
  const double PHI1 = -150.0;
  const double PSI1 =  120.0;
  gProtein.clear();
  for (int j = 0, aa = ALA; j < 20; aa++, j++)
  {
    gProtein.appendRes(aa);
    if (PRO == aa)
      gProtein.rotateProlinePhi(j, -60);
    else
      gProtein.rotatePhi(j, PHI1);
    gProtein.rotatePsi(j, PSI1);
  }
  gProtein.blockNTerminal();
  gProtein.blockCTerminal();
}



void load_protein(string name, Protein& gProtein)
{
  PdbReader pdb(name);
  if (!pdb.isFileLoaded()) return;
  pdb.loadChain(gProtein);

  if (gProtein.nRes()==0) return;

  for (int i = 0; i < gProtein.nRes(); i++)
    gProtein.buildAllHydrogens(i);

  string s1 = gProtein.getName();
  while ( (s1.length() > 0) && (s1[s1.length()-1] == ' ') )
    s1.erase(s1.length()-1, 1);

  string s = "[" + gProtein.getPdbCode() + "]  " +  s1;
  if (s == "[]  ") 
    s = name;
  s += " - The Ramachandran Plot Explorer";
  set_window_title(s.c_str());
}



void open_file()
{
  static char save_string[600] = "";
  if (file_select_ex("Open", save_string, "pdb", 500, 600, 200) != 0)
  {
    load_protein(save_string, gProtein);
    if (gProtein.isEmpty())
      build_default_protein();
    gpProteinDisplay->initializeFocus();
  }
}


void save_file()
{
  static char save_string[600] = "";

  if (save_string[0] == 0) 
    strcpy(save_string, "peptide.pdb");

  if (file_select_ex("Save pdb file", save_string, "pdb", 500, 600, 200) != 0)
  {
    PdbWriter writer(save_string);
    if (writer.isFileOpened())
    {
      writer.writeChain(gProtein, gpProteinDisplay->isHydrogen);
      string s = save_string;
      s += " - The Ramachandran Plot Explorer";
      set_window_title(s.c_str());
    }
  }
}


void clear_measure()
{
  gpProteinDisplay->clearMeasure();
}



void clear_anchor()
{
  gpProteinDisplay->clearAnchor();
}



void set_colors()
{
  PALETTE pal;

  for (int c = 1; c < 256; c++)
  {
     pal[c].r = default_palette[c].r;
     pal[c].b = default_palette[c].b;
     pal[c].g = default_palette[c].g;
  }
  pal[BLACK       ].r =       pal[BLACK       ].g =       pal[BLACK       ].b = 0;
  pal[RED         ].r = 255;  pal[RED         ].g =       pal[RED         ].b = 0;
  pal[BLUE        ].b = 255;  pal[BLUE        ].r =       pal[BLUE        ].g = 20;
  pal[YELLOW      ].r =       pal[YELLOW      ].g = 245;  pal[YELLOW      ].b = 0;
  pal[GREEN       ].g = 255;  pal[GREEN       ].r =       pal[GREEN       ].b = 0;
  pal[DARK_GREEN  ].g = 20;   pal[DARK_GREEN  ].r =       pal[DARK_GREEN  ].b = 0;
  pal[LIGHT_BLUE  ].b = 255;  pal[LIGHT_BLUE  ].r =       pal[LIGHT_BLUE  ].g = 240;
  pal[PURPLE      ].g = 0;    pal[PURPLE      ].r =       pal[PURPLE      ].b = 255;
  pal[ORANGE      ].r = 255;  pal[ORANGE      ].b = 0;    pal[ORANGE      ].g = 150;
  pal[LIGHT_ORANGE].r = 255;  pal[LIGHT_ORANGE].b = 150;  pal[LIGHT_ORANGE].g = 180;

  // set up a grey-scale in the top half of the palette
  for (int c = 128; c < 256; c++)
     pal[c].r = pal[c].g = pal[c].b = (c-128)*2;

  set_palette(pal);
}



void initialize_widgets()
{
  const int BUTTON_W = 120; 

  gpProteinDisplay = new ProteinDisplay(&gProtein, 0, 0, SCREEN_W, SCREEN_H);
  gpProteinDisplay->initializeFocus();

  gWidgetList.push_back(gpProteinDisplay);

  gWidgetList.push_back(new RamaBox    (gpProteinDisplay, 15,  SCREEN_H - 200,        170,                     170));
  gWidgetList.push_back(new ChiBox     (gpProteinDisplay, 15,  SCREEN_H - 210 - 4*16, 170,                     16) );
  gWidgetList.push_back(new FocusBox   (gpProteinDisplay, 15,  10 + 10 + 14,          14,                      6)  );
  gWidgetList.push_back(new SequenceBox(gpProteinDisplay, 15,  10 + 10,               14,                      14) );
  gWidgetList.push_back(new ScrollBar  (gpProteinDisplay, 15,  10,                    14 * SCROLLBAR_SIZE + 1, 10) );
  gWidgetList.push_back(new LoopBox    (gpProteinDisplay, 195, SCREEN_H - 45,        14,                      14) );

  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 15,        BUTTON_W, 17, &gIsAnimate,               "animate",       reset_point_of_view) );
  
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 0*20, BUTTON_W, 17, &gpProteinDisplay->isCentered,  "center",        reset_point_of_view) );
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 1*20, BUTTON_W, 17, &gpProteinDisplay->isHbond,     "H-bond",        reset_point_of_view) );
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 2*20, BUTTON_W, 17, &gpProteinDisplay->isSolid,     "solid",         reset_point_of_view) );
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 3*20, BUTTON_W, 17, &gpProteinDisplay->isSidechain, "sidechain",     reset_point_of_view) );
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 4*20, BUTTON_W, 17, &gpProteinDisplay->isPeptide,   "peptide",       reset_point_of_view) );
  gWidgetList.push_back(new ToggleButton(SCREEN_W - BUTTON_W - 5, 50 + 5*20, BUTTON_W, 17, &gpProteinDisplay->isHydrogen,  "hydrogen",      reset_point_of_view) );

  gWidgetList.push_back(new Button(      SCREEN_W - BUTTON_W - 5, 190,        BUTTON_W, 17, "save pdb",  save_file) );
  gWidgetList.push_back(new Button(      SCREEN_W - BUTTON_W - 5, 190 + 1*20, BUTTON_W, 17, "open pdb",  open_file) );

  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 250 + 0*16, BUTTON_W, 17, &gpProteinDisplay->pickState, PICK_RESIDUE,  "pick res",        reset_point_of_view) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 250 + 1*16, BUTTON_W, 17, &gpProteinDisplay->pickState, PICK_DISTANCE, "pick measure",    reset_point_of_view) );
  gWidgetList.push_back(new Button(       SCREEN_W - BUTTON_W - 5, 250 + 2*16, BUTTON_W, 17, "clear measure",  clear_measure) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 250 + 3*16, BUTTON_W, 17, &gpProteinDisplay->pickState, PICK_ANCHOR, "pick anchor",    reset_point_of_view) );
  gWidgetList.push_back(new Button(       SCREEN_W - BUTTON_W - 5, 250 + 4*16, BUTTON_W, 17, "clear anchor",  clear_anchor) );

  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 360 + 0*16, BUTTON_W, 17, &gpProteinDisplay->showState, SHOW_ALL,      "show all",        reset_point_of_view) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 360 + 1*16, BUTTON_W, 17, &gpProteinDisplay->showState, SHOW_SEGMENT,  "show scrollbar",  reset_point_of_view) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 360 + 2*16, BUTTON_W, 17, &gpProteinDisplay->showState, SHOW_SPHERE,   "show sphere",     reset_point_of_view) );

  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 450 + 0*16, BUTTON_W, 17, &gpProteinDisplay->stereoState, STEREO_NONE,         "no stereo",    reset_point_of_view) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 450 + 1*16, BUTTON_W, 17, &gpProteinDisplay->stereoState, STEREO_CROSSEYE,     "cross-eye",    reset_point_of_view) );
  gWidgetList.push_back(new StateSelector(SCREEN_W - BUTTON_W - 5, 450 + 2*16, BUTTON_W, 17, &gpProteinDisplay->stereoState, STEREO_WALLEYE,      "wall-eye",     reset_point_of_view) );

  gWidgetList.push_back(new Button(SCREEN_W - BUTTON_W - 5, SCREEN_H - 40, BUTTON_W, 17, "exit",  handle_close) );
}



int main(int argc, char **argv)
{
  allegro_init();
  install_keyboard();
  install_mouse();
  install_timer();

  if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 800, 600, 0, 0) != 0)
    if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0) != 0)
      if (set_gfx_mode(GFX_AUTODETECT, 1024, 768, 0, 0) != 0)
      {
        set_gfx_mode(GFX_TEXTSTRING, 0, 0, 0, 0);
        allegro_message("Error setting graphics mode\n%s\n", allegro_error);
        exit(1);
      }

  set_colors();

  set_window_close_hook(handle_close);
 
  set_window_title("The Ramachandran Plot Explorer");

  if (argc >= 2)
    load_protein(argv[1], gProtein);
  if (gProtein.isEmpty())
    build_default_protein();

  initialize_widgets();

  BITMAP* buffer = create_bitmap(SCREEN_W, SCREEN_H);
  int last_retrace_count = retrace_count;

  while (true)
  {
    // animate the gProtein
    while (last_retrace_count < retrace_count)
    {
      if (gIsAnimate) animate_protein();
      last_retrace_count++;
    }
    if (gIsAnimate)
    {
      gProtein.rebuild();
      reset_point_of_view();
    }

    // draw the screen
    gWidgetList.draw(buffer);
    draw_mouse(buffer);
    stringout_right(buffer, font, VERSION_STRING, SCREEN_W - 15, SCREEN_H - 12, WHITE);

    vsync();
    blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

    if (gIsProgramEnd) 
      break;

    gWidgetList.handleMouse();

    yield_timeslice();
  }

  // clean view memory
  destroy_bitmap(buffer);

  return 0;
}

END_OF_MAIN();

